/*BEGIN CMVC */
//*************************************************************************
//
//  Workfile: servlet/sample/AuditTrailHandler.java, v3-java, fmv3
//  Last update: 05/05/25 16:15:39
//  SCCS path, id: /home/flowmark/vc/0/3/5/5/s.33 1.14
//
//************************************************************************/
/*END CMVC */
/*BEGIN COPYRIGHT */
//*****************************************************************************
//
//  Licensed Materials - Property of IBM
//  5697-FM3 5655-A96
//  (C) Copyright IBM Corp. 1999, 2008
//  US Government Users Restricted Rights - Use, duplication or
//  disclosure restricted by GSA ADP Schedule Contract
//  with IBM Corp.
//
//****************************************************************************/
//END COPYRIGHT */
/*BEGIN DISCLAIMER */
//*************************************************************************
//  DISCLAIMER
//  ----------
//  This material contains programming source code for your consideration.
//  These examples have not been thoroughly tested under all conditions.
//  IBM, therefore, cannot guarantee or imply reliability, serviceability,
//  or function of these programs.
//  ALL PROGRAMS CONTAINED HEREIN ARE PROVIDED TO YOU "AS IS", WITHOUT ANY
//  WARRANTIES (EXPRESS OR IMPLIED) OR SUPPORT WHATSOEVER, INCLUDING BUT
//  NOT LIMITED TO ANY IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS
//  FOR A PARTICULAR PURPOSE.
//************************************************************************/
/*END DISCLAIMER */

package com.ibm.workflow.servlet.sample;

import java.util.*;
import java.sql.*;
import java.io.*;
import javax.servlet.http.*;

import javax.xml.transform.*;
import javax.xml.transform.stream.*;

import com.ibm.workflow.api.*;
import com.ibm.workflow.api.ProcessTemplatePackage.AuditSetting;
import com.ibm.workflow.servlet.client.*;

/**
* <p>This CommandHandler accesses the IBM WebSphere MQ Workflow audit trail using JDBC
* and returns the event data in XML format. This data is then optionally
* formatted using XSL style sheets.</p>
* <p>This sample also demonstrates how to use the {@link Command} class.</p>
* <p>For details on how to run this sample, see the
* <a href="../../../../../../samples.html#audit">Samples</a> section.</p>
*/

public class AuditTrailHandler extends GenericCommandHandler
{
  /**
  * Returns information about the handler, such as author, version,
  * and copyright.
  * @return a String containing handler information
  */

  public String getHandlerInfo() { return "MQWF Sample - Audit Trail Handler V2.1"; }

  /**
  * During initialization, the handler opens a JDBC connection to the MQWF
  * runtime database (either on the local host or on a remote host [depending
  * on the presence of the 'DB2Server' property in the configuration file]).
  * This connection is cached. Then two SQL statements are prepared which will
  * be used to retrieve the audit data for either a particular process instance
  * or all the data available for a particular activity (across all process
  * instances). If configured, the corresponding style sheets are loaded and
  * preprocessed.
  * @param config  The servlet's configuration and initialization parameters
  */

  public void init(Config cfg)
  {
    super.init(cfg);

    String DB2Server    = cfg.getParameter("AuditTrail", "DB2Server"              );
    String DB2User      = cfg.getParameter("AuditTrail", "DB2User",     "db2admin");
    String DB2Password  = cfg.getParameter("AuditTrail", "DB2Password", "db2admin");
    String databaseName = cfg.getParameter("AuditTrail", "Database",    "FMCDB"   );

    try
    {
      if (DB2Server == null)
      {
        Class.forName("COM.ibm.db2.jdbc.app.DB2Driver");

        DB2Server = "";
      }
      else
      {
        Class.forName("COM.ibm.db2.jdbc.net.DB2Driver");

        DB2Server = "//" + DB2Server + "/";
      }

      writelog("      DB2Server:     " + DB2Server,    false);
      writelog("      DB2User:       " + DB2User,      false);
      writelog("      Database:      " + databaseName, false);

      if (DB2User != null)
        conn = DriverManager.getConnection("jdbc:db2:" + DB2Server + databaseName, DB2User, DB2Password);
      else
        conn = DriverManager.getConnection("jdbc:db2:" + DB2Server + databaseName);

      stmtReport = conn.prepareStatement("SELECT * FROM FMC.AUDIT_TRAIL WHERE PROCESS_ID=?");

      stmtStats  = conn.prepareStatement("SELECT * FROM FMC.AUDIT_TRAIL " +
                                         "WHERE ACTIVITY_NAME=? AND PROC_TEMPL_NAME=? " +
                                         "AND EVENT >= 21021 AND EVENT <= 21023 " +
                                         "ORDER BY USER_NAME ASC");

      audit  = cfg.getParameter("AuditTrail", "StyleSheetPI");
      stats  = cfg.getParameter("AuditTrail", "StyleSheetWI");
      reload = cfg.getParameter("AuditTrail", "Reload", false);

      if (audit != null || stats != null)
      {
        try
        {
          tFactory   = TransformerFactory.newInstance();
          css        = getURL("webclientstyle.css");
          imgLogoff  = getURL("images/navigation/logoff.gif");
          imgRefresh = getURL("images/navigation/refresh.gif");

          if (audit != null)
          {
            tAudit   = tFactory.newTemplates(new StreamSource(new File(audit = getContext().getRealPath(audit))));
            imgAudit = getURL("images/navigation/back_instlist.gif");

            writelog("      StyleSheetPI:  " + audit, false);
          }

          if (stats != null)
          {
            tStats   = tFactory.newTemplates(new StreamSource(new File(stats = getContext().getRealPath(stats))));
            imgStats = getURL("images/navigation/back_worklist.gif");

            writelog("      StyleSheetWI:  " + stats, false);
          }
        }
        catch (NoClassDefFoundError xcpt) { writelog("      TrAX (XSL) not available: " + xcpt.toString(), true); }
      }
    }
    catch(ClassNotFoundException xcpt)            { writelog("     init: DB2 JDBC driver not found.", true ); }
    catch(SQLException xcpt)                      { writelog("     init:", xcpt); }
    catch(TransformerConfigurationException xcpt) { writelog("     init:", xcpt); }
  }

  /**
  * When the handler is unloaded, the cached JDBC connection is closed.
  */

  public void destroy()
  {
    try
    {
      if (conn != null) conn.close();
    }
    catch(SQLException xcpt) { writelog("destroy:", xcpt); }
  }

  /**
  * The trigger tags for the handler. Those URL's will be inserted in the corresponding lists (and show up there as the
  * icon which is referenced in the URL. Note that the function returns an empty string if no JDBC connection could be
  * established (which will prevent the display of the icons).
  * @param list Denotes the list for which the trigger tag is queried.
  * @param context The {@link RequestContext} which was passed to the {@link Viewer}
  *                calling this method.
  * @return For work lists, an HTML tag to inovke the
  *         {@link #getActivityStatistics getActivityStatistics}
  *         command; for instance lists, an HTML tag to invoke the
  *         {@link #getInstanceHistory getInstanceHistory} command.
  */

  public String getTriggerTagFor(int list, RequestContext context)
  {
    String tag = "";

    try
    {
      if      (conn == null) ;
      else if (list == SessionContext.WORKLIST)
      {
        WorkItem wi = context.getWorkItem();

        if (wi != null) tag = cmdStats.getTriggerTag(context, wi.persistentOid(), null);
      }
      else if (list == SessionContext.INSTANCELIST)
      {
        tag = cmdReport.getTriggerTag(context, context.getInstance().persistentOid(), null);
      }
    }
    catch (FmcException xcpt) { writelog("getTriggerTagFor: ", xcpt); }

    return tag;
  }

  /**
  * This custom command first retrieves the audit records for the process
  * instance whose OID matches the 'id' parameter in the HTTP-request, and
  * then transforms the data to XML. If a style sheet name is either passed
  * via the 'stylesheet' parameter or is configured, the XML data is passed
  * to it and the resulting HTML data is returned; otherwise the XML data
  * is returned.
  * @param request  The request the client has made of the servlet
  * @param response The response the servlet sends to the client
  * @return The result page to be sent to the client
  */

  public ResponsePage getInstanceHistory(HttpServletRequest request,
                                         HttpServletResponse response)
                      throws ClientException
  {
    StringBuffer xml;
    StringWriter html = null;
    long         t1   = System.currentTimeMillis();
    long         t2, t3;

    try
    {
      synchronized (this)
      {
        String oid  = request.getParameter("id");

        stmtReport.setString(1, oid);

        xml = process(stmtReport.executeQuery());

        t2 = System.currentTimeMillis();

        String ssn  = request.getParameter("stylesheet");

        if (ssn != null) // Make sure this works without a session context, so it can be called anonymously.
        {
          ApplicationContext cxt = getContext();

          html = process(null, xml,
                         tFactory.newTemplates(new StreamSource(new File(cxt.getRealPath(cxt.getDocumentRoot() + ssn)))),
                         null, null, null, null, oid);
        }
        else if (tAudit != null)
        {
          if (reload) tAudit = tFactory.newTemplates(new StreamSource(new File(audit)));

          SessionContext cxt = getSessionContext(request);

          html = process(cxt, xml, tAudit, imgAudit, "queryProcessInstances", cxt.getInstanceListOid(),
                         cmdReport.getCommand(), oid);
        }

        t3 = System.currentTimeMillis();
      }
    }
    catch(Exception xcpt) { throw new ClientException(request, xcpt); }

    writelog("AuditTrail XML processing took " + (t2-t1) + "msec, XSL processing took " + (t3-t2) + "msec", false);

    return html == null ? new ResponsePage(xml,ResponsePage.MIMETYPE_XML)
                        : new ResponsePage(html.getBuffer());
  }

  /**
  * This custom command first retrieves the audit records for all activity
  * instances whose names match the name of work item whose OID is passed via
  * the 'id' parameter of the HTTP-request. Then the audit data is transformed
  * to XML. If a style sheet is configured, the XML data is passed to it and
  * the resulting HTML data is returned; otherwise the XML data is returned.
  * @param request  The request the client has made of the servlet
  * @param response The response the servlet sends to the client
  * @return The result page to be sent to the client
  */

  public ResponsePage getActivityStatistics(HttpServletRequest request,
                                            HttpServletResponse response)
                      throws ClientException
  {
    StringBuffer xml;
    StringWriter html = null;
    long         t1   = System.currentTimeMillis();
    long         t2, t3;

    try
    {
      SessionContext context = getSessionContext(request);
      String         loid    = context.getWorkListOid();
      String         oid     = request.getParameter("id");
      WorkItem       wi      = context.getWorkItem(loid, oid);

      synchronized (this)
      {
        stmtStats.setString(1, wi.activityInstance().name());
        stmtStats.setString(2, wi.processInstance().processTemplateName());

        xml = process(stmtStats.executeQuery());

        t2 = System.currentTimeMillis();

        if (tStats != null)
        {
          if (reload) tStats =  tFactory.newTemplates(new StreamSource(new File(stats)));

          html = process(context, xml, tStats, imgStats,
                         "queryWorkItems", loid, cmdStats.getCommand(), oid);
        }

        t3 = System.currentTimeMillis();
      }
    }
    catch(Exception xcpt) { throw new ClientException(request, xcpt); }

    writelog("Statistics XML processing took " + (t2-t1) + "msec, XSL processing took " + (t3-t2) + "msec", false);

    return html == null ? new ResponsePage(xml,ResponsePage.MIMETYPE_XML)
                        : new ResponsePage(html.getBuffer());
  }

  /**
  * This static method transforms the numeric representation of an MQWF event
  * code to a human-readable string by looking up the translation in the
  * 'AuditTrail.properties' file (which allows for easy language translation).
  * This function is excercised by the 'AuditTrail.xsl' style sheet and is an
  * example of how style sheets can call back to Java code.
  * @param code The numeric audit event ID
  * @return A desription for the audit event
  */

  public static final String format(String code) { return cat.get(code); } // ... read from catalog ...

  /**
  * Given a ResultSet 'rs', this function translates the SQL data to XML
  * according to the DTD kept in member 'dtd' (which is inlined into the XML
  * data as well to allow for XML validation). Once done, the ResultSet is
  * discarded.
  */

  private final StringBuffer process(ResultSet rs) throws SQLException
  {
    StringBuffer sb = new StringBuffer(2500);

    sb.append(dtd + "<AuditTrail>");

    while(rs.next())
    {
      sb.append("<AuditTrailRecord>");

      write(sb, rs, rs.getTimestamp("CREATED"           ), "Timestamp"             );
      write(sb, rs, rs.getInt      ("EVENT"             ), "AuditEvent"            );
      write(sb, rs, rs.getString   ("PROCESS_NAME"      ), "ProcInstName"          );
      write(sb, rs, rs.getString   ("PROCESS_ID"        ), "ProcInstID"            );
      write(sb, rs, rs.getString   ("TOP_LVL_PROC_NAME" ), "ProcInstTopLevelName"  );
      write(sb, rs, rs.getString   ("TOP_LVL_PROC_ID"   ), "ProcInstTopLevelID"    );
      write(sb, rs, rs.getString   ("PARENT_PROC_NAME"  ), "ProcInstParentName"    );
      write(sb, rs, rs.getString   ("PARENT_PROC_ID"    ), "ProcInstParentID"      );
      write(sb, rs, rs.getString   ("PROC_TEMPL_NAME"   ), "ProcTemplName"         );
      write(sb, rs, rs.getTimestamp("TEMPL_VALID_FROM"  ), "ProcTemplValidFromDate");
      write(sb, rs, rs.getString   ("BLOCK_NAMES"       ), "BlockNames"            );
      write(sb, rs, rs.getString   ("USER_NAME"         ), "UserID"                );
      write(sb, rs, rs.getString   ("SECOND_USER_NAME"  ), "SecondUserID"          );
      write(sb, rs, rs.getString   ("ACTIVITY_NAME"     ), "ActivityName"          );
      write(sb, rs, rs.getInt      ("ACTIVITY_TYPE"     ), "ActivityType"          );
      write(sb, rs, rs.getInt      ("ACTIVITY_STATE"    ), "ActivityState"         );
      write(sb, rs, rs.getString   ("SECOND_ACT_NAME"   ), "SecondActivityName"    );
      write(sb, rs, rs.getString   ("COMMAND_PARAMETERS"), "CommandParameters"     );
      write(sb, rs, rs.getString   ("ASSOCIATED_OBJECT" ), "AssociatedObject"      );
      write(sb, rs, rs.getString   ("OBJECT_DESCRIPTION"), "ObjectDescription"     );
      write(sb, rs, rs.getString   ("PROGRAM_NAME"      ), "ProgramName"           );
      write(sb, rs, rs.getInt      ("ACTIVITY_RC"       ), "ProgramRc"             );

      sb.append("</AuditTrailRecord>");
    }

    sb.append( "</AuditTrail>");

    rs.close();

    return sb;
  }

  /**
  * This method takes XML data ('xml'), a compiled style sheet ('tpl'), as well
  * as a set of navigation URL's and images (imgBack, cmdBack, idBack,
  * imgRefresh and cmdRefresh) and produces HTML via the TrAX processor. The
  * URLs (which make up the navigation bar in the resulting HTML page) are
  * passed to the style sheet as xsl:param's.
  */

  private final StringWriter process(SessionContext context, StringBuffer xml,
                                     Templates tpl, String imgBack,
                                     String cmdBack, String idBack,
                                     String cmdRefresh, String idRefresh)
    throws IOException, TransformerException
  {
    StringWriter sw = new StringWriter();
    Transformer  tr = tpl.newTransformer();

    tr.setParameter("css",        css);
    tr.setParameter("imgBack",    imgBack);
    tr.setParameter("imgLogoff",  imgLogoff);
    tr.setParameter("imgRefresh", imgRefresh);

    if (context != null)
    {
      tr.setParameter("cmdBack",    context.getCommand(cmdBack, idBack, "&useCache=true"));
      try { tr.setParameter("cmdRefresh", java.net.URLDecoder.decode(context.getCommand(cmdRefresh, idRefresh))); }
      catch (Throwable ignored) { /* JDK1.2's decode() may throw a java.lang.Exception */ }
      tr.setParameter("cmdLogoff",  context.getCommand("logoff"));
    }

    tr.transform(new StreamSource(new StringReader(xml.toString())), new StreamResult(sw));

    sw.close();

    return sw;
  }

  /**
  * Write the string, integer or timestamp 'value' to the 'xml'-buffer, using
  * 'tag' element. Dont write anything if SQL returned 'IS NULL'.
  */

  private final void write(StringBuffer xml, ResultSet rs, String value, String tag) throws SQLException
  {
    if (!rs.wasNull()) write(xml, value, tag);
  }

  private final void write(StringBuffer xml, ResultSet rs, long value, String tag) throws SQLException
  {
    if (!rs.wasNull()) write(xml, Long.toString(value), tag);
  }

  private final void write(StringBuffer xml, ResultSet rs, Timestamp value, String tag) throws SQLException
  {
    if (!rs.wasNull()) write(xml, rs, value.getTime() + value.getNanos()/1000000, tag);
  }

  /**
  * Write an string to the 'xml'-buffer. The given string is 'XML-sanitized'
  * (non-ASCII and special characters [i.e. those used in XML markup] are
  * escaped). The string is wrapped using the given element tag.
  */

  private final void write(StringBuffer xml, String value, String tag)
  {
    xml.append("<" + tag + ">");
    xml.append(RequestContext.toHTMLText(value, ""));
    xml.append("</" + tag + ">");
  }

  /**
  * Convenience function to transform a relative URL to an absolute URL.
  */

  private final String getURL(String name)
  {
    return getContext().getURL("{_HTMLDir_}/" + name);
  }

  /**
  * Convenience functions to log messages and exceptions.
  */

  private final void writelog(String logEntry, boolean master)
  {
    getContext().log((master ? pre : "") + logEntry, master);
  }

  private final void writelog(String logEntry, Throwable xcpt)
  {
    getContext().log(pre + logEntry, xcpt);
  }

  private boolean            reload;
  private String             css;
  private String             audit;
  private String             stats;
  private String             imgRefresh;
  private String             imgLogoff;
  private String             imgAudit;
  private String             imgStats;
  private Connection         conn;
  private PreparedStatement  stmtReport;
  private PreparedStatement  stmtStats;
  private TransformerFactory tFactory;
  private Templates          tAudit;
  private Templates          tStats;
  // AudiTrail.properties contains the english string representations of the events recorded in the AuditTrail -> Locale is set to en
  private static com.ibm.workflow.servlet.client.CatalogImpl   cat   = new com.ibm.workflow.servlet.client.CatalogImpl( "com.ibm.workflow.catalog.servlet.sample.AuditTrail", Locale.ENGLISH);

  private static final String  pre       = "AuditTrailHandler: ";
  private static final Command cmdReport = new Command("x-getInstanceHistory",    "audittrail.gif", "Audit Trail History");
  private static final Command cmdStats  = new Command("x-getActivityStatistics", "audittrail.gif", "Activity Statistics");
  private static final String  dtd       = "<?xml version='1.0' encoding='ISO-8859-1'?>\n"
                                         + "<!DOCTYPE AuditTrail [\n"
                                         + "<!ELEMENT AuditTrail (AuditTrailRecord*)>\n"
                                         + "<!ELEMENT AuditTrailRecord\n"
                                         + " (Timestamp,\n"
                                         + "  AuditEvent,\n"
                                         + "  ProcInstName,\n"
                                         + "  ProcInstID,\n"
                                         + "  ProcInstTopLevelName,\n"
                                         + "  ProcInstTopLevelID,\n"
                                         + "  ProcInstParentName?,\n"
                                         + "  ProcInstParentID?,\n"
                                         + "  ProcTemplName,\n"
                                         + "  ProcTemplValidFromDate?,\n"
                                         + "  BlockNames?,\n"
                                         + "  UserID?,\n"
                                         + "  SecondUserID?,\n"
                                         + "  ActivityName?,\n"
                                         + "  ActivityType?,\n"
                                         + "  ActivityState?,\n"
                                         + "  SecondActivityName?,\n"
                                         + "  CommandParameters?,\n"
                                         + "  AssociatedObject?,\n"
                                         + "  ObjectDescription?,\n"
                                         + "  ProgramName?,\n"
                                         + "  ProgramRc?)>\n"
                                         + "<!ELEMENT Timestamp              (#PCDATA)>\n"
                                         + "<!ELEMENT AuditEvent             (#PCDATA)>\n"
                                         + "<!ELEMENT ProcInstName           (#PCDATA)>\n"
                                         + "<!ELEMENT ProcInstID             (#PCDATA)>\n"
                                         + "<!ELEMENT ProcInstTopLevelName   (#PCDATA)>\n"
                                         + "<!ELEMENT ProcInstTopLevelID     (#PCDATA)>\n"
                                         + "<!ELEMENT ProcInstParentName     (#PCDATA)>\n"
                                         + "<!ELEMENT ProcInstParentID       (#PCDATA)>\n"
                                         + "<!ELEMENT ProcTemplName          (#PCDATA)>\n"
                                         + "<!ELEMENT ProcTemplValidFromDate (#PCDATA)>\n"
                                         + "<!ELEMENT BlockNames             (#PCDATA)>\n"
                                         + "<!ELEMENT UserID                 (#PCDATA)>\n"
                                         + "<!ELEMENT SecondUserID           (#PCDATA)>\n"
                                         + "<!ELEMENT ActivityName           (#PCDATA)>\n"
                                         + "<!ELEMENT ActivityType           (#PCDATA)>\n"
                                         + "<!ELEMENT ActivityState          (#PCDATA)>\n"
                                         + "<!ELEMENT SecondActivityName     (#PCDATA)>\n"
                                         + "<!ELEMENT CommandParameters      (#PCDATA)>\n"
                                         + "<!ELEMENT AssociatedObject       (#PCDATA)>\n"
                                         + "<!ELEMENT ObjectDescription      (#PCDATA)>\n"
                                         + "<!ELEMENT ProgramName            (#PCDATA)>\n"
                                         + "<!ELEMENT ProgramRc              (#PCDATA)>\n"
                                         + "]>\n";
}
