/*BEGIN CMVC */
//*************************************************************************
//
//  Workfile: servlet/sample/BasicAuthenticationHandler.java, v3-java, fmv3
//  Last update: 05/05/25 16:15:34
//  SCCS path, id: /home/flowmark/vc/0/3/7/1/s.09 1.10
//
//************************************************************************/
/*END CMVC */
/*BEGIN COPYRIGHT */
//*****************************************************************************
//
//  Licensed Materials - Property of IBM
//  5697-FM3 5655-A96
//  (C) Copyright IBM Corp. 1999, 2008
//  US Government Users Restricted Rights - Use, duplication or
//  disclosure restricted by GSA ADP Schedule Contract
//  with IBM Corp.
//
//****************************************************************************/
//END COPYRIGHT */
/*BEGIN DISCLAIMER */
//*************************************************************************
//  DISCLAIMER
//  ----------
//  This material contains programming source code for your consideration.
//  These examples have not been thoroughly tested under all conditions.
//  IBM, therefore, cannot guarantee or imply reliability, serviceability,
//  or function of these programs.
//  ALL PROGRAMS CONTAINED HEREIN ARE PROVIDED TO YOU "AS IS", WITHOUT ANY
//  WARRANTIES (EXPRESS OR IMPLIED) OR SUPPORT WHATSOEVER, INCLUDING BUT
//  NOT LIMITED TO ANY IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS
//  FOR A PARTICULAR PURPOSE.
//************************************************************************/
/*END DISCLAIMER */

package com.ibm.workflow.servlet.sample;

import javax.servlet.http.*;

import com.ibm.workflow.servlet.client.*;
import com.ibm.workflow.api.*;
import com.ibm.workflow.api.ServicePackage.*;

/**
* <p>This CommandHandler extracts the <a href="http://rfc.net/rfc2068.html">
* HTTP/1.1 Basic Authentication</a> header
* from a request and uses it to logon on to IBM WebSphere MQ Workflow, if present.
* It can also create credentials for Servlet V2.2 form based
* authentication.</p>
* <p>Since no user ID mapping currently is performed, the user ID in the
* Authentication header must be a valid Workflow user ID.</p>
* <p>For details on how to run this sample, see the
* <a href="../../../../../../samples.html#bauth">Samples</a> section.</p>
*/
public class BasicAuthenticationHandler extends GenericCommandHandler
{
  /**
  * Returns information about the handler, such as author, version,
  * and copyright.
  * @return a String containing handler information
  */
  public String getHandlerInfo()
  {
    return "MQWF Sample - Basic HTTP/1.1 Authentication V1.2";
  }

  /**
  * Checks if the request contains HTTP/1.1 Basic Authentication credentials.
  * If so, they are used for the authenication exit, that is, the request
  * parameters (<code>userID</code>, ...) passed to the
  * {@link BuiltinHandler#logon BuiltinHandler.logon()}
  * command will be ignored.
  * @param request The logon request sent by the client
  * @return <code>null</code> if no Authentication Header is present. If the
  *         the Authentication header is present, the
  *         <code>ExecutionService.logon4()</code> method will be called and
  *         an authentication exit must have been installed on the Workflow
  *         server.
  * @exception ClientException If an error occurred.
  */
  public byte[] getCredentials(HttpServletRequest request)
                throws ClientException
  {
    String credentials = null;
    String authType = request.getAuthType();

    /* HTTP/1.1 specs (RFC2068) - 11.1 Basic Authentication Scheme         */
    /* -----------------------------------------------------------         */
    /* To receive authorization, the client sends the userid and password, */
    /* separated by a single colon (":") character, within a base64        */
    /* encoded string in the credentials.                                  */

    if (credentials == null

       )
    {
      String auth = request.getHeader("Authorization");

      if (auth != null && auth.startsWith("Basic "))
      {
        auth = decodeBase64(auth.substring(6).trim());

        if (auth.indexOf(':') > 0)
        {
          credentials = auth;

          String userID = auth.substring(0, auth.indexOf(':'));
          getContext().log("BasicAuthenticationHandler: Using basic authentication for " +
                           userID, false);
        }
      }
    }

    // Servlet V2.2 specs - 11.5.3 Form Based Authentication
    if (credentials == null && "FORM".equalsIgnoreCase(authType))
    {
      // The user ID entered in the 'login' form
      String userID = request.getRemoteUser();

      if (userID != null)
      {
        // With form-based authorization the password is neither sent in
        // the HTTP header nor is it accessible programmatically.
        // Therefore, we assume the hard-coded value 'password'. A sensible
        // implementation is only possible with IBM WebSphere MQ Workflow's
        // authentication exit.
        credentials = userID + ":password";

        getContext().log("BasicAuthenticationHandler: Using form based authentication for " +
                         userID, false);
      }
    }

    getContext().log("BasicAuthenticationHandler: Authorization type = " +
                     authType + ", credentials = " +
                     (credentials == null ? "null" : (credentials.length() + " bytes")), false);

    if (credentials != null)
    {
      try
      {
        return credentials.getBytes("UTF8");
      }
      catch (java.io.UnsupportedEncodingException ignored) { }
    }

    return null;
  }

  /* MIME part one (RFC2045) - 6.8 Base64 Content-Transfer-Encoding      */
  /* --------------------------------------------------------------      */
  private static int base64[] =
  {
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 62, 64, 64, 64, 63,
    52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 64, 64, 64, 64, 64, 64,
    64,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14,
    15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 64, 64, 64, 64, 64,
    64, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40,
    41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 64, 64, 64, 64, 64,
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64
  };

  /**
  * Decodes a base64 encoded String. Invalid characters simply are ignored.
  * @param str a base64-encoded String
  * @return the decoded String
  */
  private static String decodeBase64(String str)
  {
    StringBuffer rc = new StringBuffer();
    int accu = 0, shift = 0, len = str.length();

    for (int i = 0; i < len; ++i)
    {
      int v = base64[str.charAt(i) & 0xff];

      if (v < 64)
      {
        accu = (accu << 6) | v;

        if (shift < 2)
          shift += 6;
        else
        {
          shift -= 2;
          rc.append((char) ((accu >> shift) & 0xff));
        }
      }
    }

    return rc.toString();
  }
}
