/*BEGIN CMVC */
//*************************************************************************
//
//  Workfile: servlet/sample/CommandHandlerAdapter.java, v3-java, fmv3
//  Last update: 05/05/25 16:15:39
//  SCCS path, id: /home/flowmark/vc/0/3/5/5/s.38 1.10
//
//************************************************************************/
/*END CMVC */
/*BEGIN COPYRIGHT */
//*****************************************************************************
//
//  Licensed Materials - Property of IBM
//  5697-FM3 5655-A96
//  (C) Copyright IBM Corp. 1999, 2008
//  US Government Users Restricted Rights - Use, duplication or
//  disclosure restricted by GSA ADP Schedule Contract
//  with IBM Corp.
//
//****************************************************************************/
//END COPYRIGHT */
/*BEGIN DISCLAIMER */
//*************************************************************************
//  DISCLAIMER
//  ----------
//  This material contains programming source code for your consideration.
//  These examples have not been thoroughly tested under all conditions.
//  IBM, therefore, cannot guarantee or imply reliability, serviceability,
//  or function of these programs.
//  ALL PROGRAMS CONTAINED HEREIN ARE PROVIDED TO YOU "AS IS", WITHOUT ANY
//  WARRANTIES (EXPRESS OR IMPLIED) OR SUPPORT WHATSOEVER, INCLUDING BUT
//  NOT LIMITED TO ANY IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS
//  FOR A PARTICULAR PURPOSE.
//************************************************************************/
/*END DISCLAIMER */

// *********************************************************************
// *                                                                   *
// * IBM WebSphere MQ Workflow CommandHandler Adapter Example                  *
// *                                                                   *
// *********************************************************************

package com.ibm.workflow.servlet.sample;

import java.util.*;
import javax.servlet.http.*;

import com.ibm.workflow.api.ExecutionService;
import com.ibm.workflow.api.FmcException;
import com.ibm.workflow.api.ServicePackage.*;
import com.ibm.workflow.servlet.client.*;

/**
* <p>CommandHandlerAdapter acts as a CommandHandler which aggregates other
* CommandHandlers. This will be useful if multiple CommandHandlers should be
* used together in a single MQWF Web Client servlet.</p>
* <p>This sample also demonstrates how to use the {@link Config} class
* to read custom settings from <code>WebClient.properties</code>.</p>
* <p>For details on how to run this sample, see the
* <a href="../../../../../../samples.html#adapter">Samples</a> section.</p>
*/
public class CommandHandlerAdapter implements CommandHandler
{
  /**
  * Returns information about the handler, such as author, version,
  * and copyright.
  * @return a String containing handler information
  */
  public final String getHandlerInfo()
  {
    StringBuffer info = new StringBuffer("\r\n\tMQWF Sample - CommandHandlerAdapter V1.1");

    for (int i = 0; i < cnt; i++)
      info.append("\r\n\t\t").append(ch[i].getHandlerInfo());

    return info.toString();
  }

  /**
  * This method can be called by {@link Viewer}s to obtain a HTML tag that
  * triggers user-defined commands.
  * @param list    Denotes the list for which the trigger tag is queried.
  * @param context The {@link RequestContext} which was passed to the {@link Viewer}
  *                calling this method.
  * @return A string containing HTML tags for the icons to be displayed on
  * <code>list</code>.
  */
  public final String getTriggerTagFor(int list, RequestContext context)
  {
    StringBuffer tag = new StringBuffer();

    // concatenate all trigger tags
    for (int i = 0; i < cnt; i++)
      tag.append(ch[i].getTriggerTagFor(list, context));

    return tag.toString();
  }

  /**
  * This method is called by the Web Client in its servlet's
  * <code>init()</code> method. Here the <code>[CommandHandlerAdapter]</code>
  * section of the <code>WebClient.properties</code> file is read and the
  * CommandHandlers registered there are created.
  * @param config  The servlet's configuration and initialization parameters
  */
  public final void init(Config cfg)
  {
    Hashtable ht = cfg.getParameters("CommandHandlerAdapter");

    // create all registered command handlers
    for (Enumeration e = ht.keys(); e.hasMoreElements(); )
    {
      String clazz = (String)e.nextElement();

      try
      {
        CommandHandler obj = (CommandHandler)Class.forName(clazz).newInstance();
        ch[Integer.parseInt((String)ht.get(clazz))] = obj;
      }
      catch (Throwable xcpt)
      {
        cfg.getContext().log("Error loading " + clazz + ":", xcpt);
      }
    }

    // initialize the loaded command handlers
    for (int i = 0; i < ch.length; i++)
    {
      CommandHandler c = ch[i];

      if (c != null)
      {
        c.init(cfg);
        ch[i] = null;
        ch[cnt++] = c;
      }
    }
  }

  /**
  * This method is called by the Web Client in its servlet's
  * <code>destroy()</code> method.
  * This call just is dispatched to the registered command handlers.
  */
  public final void destroy()
  {
    for (int i = 0; i < cnt; i++)
      ch[i].destroy();
  }

  /**
  * This method is called whenever a new user has logged on.
  * This call just is dispatched to the registered command handlers.
  * @param context The {@link RequestContext} for the newly added session.
  */
  public final void onLogon(RequestContext context)
  {
    for (int i = 0; i < cnt; i++)
      ch[i].onLogon(context);
  }

  /**
  * This method is called whenever a user has logged off.
  * This call just is dispatched to the registered command handlers.
  * @param context The {@link SessionContext} for the session that will
  *                be removed.
  */
  public final void onLogoff(SessionContext context)
  {
    for (int i = 0; i < cnt; i++)
      ch[i].onLogoff(context);
  }

  /**
  * This method is called by the {@link BuiltinHandler} as a pre-logon hook
  * to enable third party authentication.
  * @param request The logon request sent by the client
  * @return <code>null</code> if normal logon processing is required or the
  *         credentials extracted from the request that should be used to
  *         log on to Workflow.
  * @exception ClientException If an error occurred.
  */
  public final byte[] getCredentials(HttpServletRequest request)
                      throws ClientException
  {
    byte[] credentials = null;

    for (int i = 0; i < cnt; i++)
    {
      credentials = ch[i].getCredentials(request);

      if (credentials != null)
        break;
    }

    return credentials;
  }

  /**
  * This method is called every time a custom command is to be executed.
  * It returns the {@link ResponsePage} representing the results of
  * <code>command</code> or <code>null</code> if the command was not recognized.
  * @param command  The name of the command (the value of the
  *                 <code>command</code> request parameter)
  * @param request  The request the client has made of the servlet
  * @param response The response the servlet sends to the client
  * @return The page representing the results of <code>command</code>
  * @exception com.ibm.workflow.servlet.client.ClientException If an error occurred.
  */
  public final ResponsePage execute(String command, HttpServletRequest request,
                                    HttpServletResponse response)
                            throws ClientException
  {
    if (command.startsWith("x-"))
      command = command.substring(2);

    // loop all handlers in the defined order until the command has been
    // handled.
    for (int i = 0; i < cnt; ++i)
    {
      try
      {
        ResponsePage result = ch[i].execute(command, request, response);

        if (result != null)
          return result;
      }
      catch (NoSuchMethodException ignored)
      {
        // simply try next handler...
      }
    }

    return null;
  }

  /** CommandHandlers in configuration order */
  private CommandHandler[] ch  = new CommandHandler[16];
  /** number of CommandHandlers */
  private int              cnt = 0;
}
