/*BEGIN CMVC */
//*************************************************************************
//
//  Workfile: servlet/client/DefaultViewer.java, v3-java, fmv3
//  Last update: 05/05/04 11:24:21
//  SCCS path, id: /home/flowmark/vc/0/3/3/1/s.83 1.30
//
//************************************************************************/
/*END CMVC */
/*BEGIN COPYRIGHT */
//*****************************************************************************
//
//  Licensed Materials - Property of IBM
//  5697-FM3 5655-A96
//  (C) Copyright IBM Corp. 1999, 2008
//  US Government Users Restricted Rights - Use, duplication or
//  disclosure restricted by GSA ADP Schedule Contract
//  with IBM Corp.
//
//****************************************************************************/
//END COPYRIGHT */
/*BEGIN DISCLAIMER */
//*************************************************************************
//  DISCLAIMER
//  ----------
//  This material contains programming source code for your consideration.
//  These examples have not been thoroughly tested under all conditions.
//  IBM, therefore, cannot guarantee or imply reliability, serviceability,
//  or function of these programs.
//  ALL PROGRAMS CONTAINED HEREIN ARE PROVIDED TO YOU "AS IS", WITHOUT ANY
//  WARRANTIES (EXPRESS OR IMPLIED) OR SUPPORT WHATSOEVER, INCLUDING BUT
//  NOT LIMITED TO ANY IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS
//  FOR A PARTICULAR PURPOSE.
//************************************************************************/
/*END DISCLAIMER */

package com.ibm.workflow.servlet.client;

import java.util.*;
import java.io.IOException;
import javax.servlet.http.HttpServletRequest;

import com.ibm.workflow.api.*;

/**
* This class provides the default implementation of the {@link Viewer}
* interface. Extend this class if you want to change the appearance of only
* some response pages.<br>
* For some commands, the response pages are created dynamically (such as
* {@link #queryWorkItemsResponse queryWorkItemsResponse()}). For others, a
* static HTML file is loaded (for example,
* {@link #showTemplatePropertiesResponse showTemplatePropertiesResponse()}).
* There are also methods that simply forward to other methods - for example,
* {@link #createInstanceResponse createInstanceResponse()} refers to
* {@link #queryProcessTemplatesResponse queryProcessTemplatesResponse()} so that
* when a new process instance has been created from the template list, the
* template list is displayed again.
*/
public class DefaultViewer implements Viewer, java.io.Serializable
{
  /**
  * Initializes this class according to the <code>[DefaultViewer]</code>
  * section in the
  * <a href="../../../../../../properties.html">servlet configuration file</a>.
  * @param config the servlet configuration.
  */
  public void init(Config config)
  {
    enableTemplatelists =
      config.getParameter("DefaultViewer", "EnableTemplatelists", true);
    enableInstancelists =
      config.getParameter("DefaultViewer", "EnableInstancelists", true);
    enableWorklists =
      config.getParameter("DefaultViewer", "EnableWorklists", true);
    enableLists =
      config.getParameter("DefaultViewer", "EnableLists", true);

    autoRefresh = config.getParameter("DefaultViewer", "AutoRefresh", false);

    pageAfterLogoff = config.getParameter("DefaultViewer", "PageAfterLogoff",
                                          "{_HTMLDir_}/Logon.html");
    pageAfterLogoff = config.getContext().getURL(pageAfterLogoff);

    generateForms = config.getParameter("DefaultViewer", "GenerateForms", true);
  }

  /**
  * Creates the page that is shown in case of an error.
  * This page will be loaded from <code>forms/Error.html</code> and allows
  * the following substitution variables:
  * <table border="1" cellspacing="0" cellpadding="4">
  *   <tr>
  *     <th>Name</th>
  *     <th>Replacement</th>
  *   </tr>
  *   <tr>
  *     <td>{_HTMLDir_}</td>
  *     <td>The root URI of the servlet's web application, for example,
  *         <code>/MQWFClient</code>.</td>
  *   </tr>
  *   <tr>
  *     <td>{_MQWFServlet_}</td>
  *     <td>The servlet's name, for example,
  *         <code>/MQWFClient/servlet/Main</code>.</td>
  *   </tr>
  *   <tr>
  *     <td>{_Title_}</td>
  *     <td>The page title.</td>
  *   </tr>
  *   <tr>
  *     <td>{_Message_}</td>
  *     <td>The error message.</td>
  *   </tr>
  *   <tr>
  *     <td>{_StackTrace_}</td>
  *     <td>The Java stack trace.</td>
  *   </tr>
  * </table>
  * If the HTML template file cannot be loaded, an error page will be
  * dynamically created.
  * @param context The request context.
  * @return The response page.
  */
  public ResponsePage errorResponse(RequestContext context)
  {
    ClientException error = context.getException();
    HttpServletRequest request = error.getRequest();

    if (request != null)
    {
      String command = request.getParameter("command");

      if (command == null)
        return logoffResponse(context);

      context.log("Error(" + request.getParameter("command") + "): " +
                  Util.getMessage(error), true);

      Throwable nested = error.getNestedException();

      if (nested instanceof FmcException)
      {
        FmcException xcpt = (FmcException) nested;

        if (xcpt.rc == FmcException.FMC_ERROR_COMMUNICATION &&
            "logon".equals(command))
        {
          // communication error during logon -  MQWF server not started?
          error = new ClientException(request, context.get("Error.NoConnection"));
          context.setException(error);
        }
        else if (xcpt.rc == FmcException.FMC_ERROR_QRY_RESULT_TOO_LARGE)
        {
          // show list properties to allow changing the threshold
          try
          {
            context.setKindOfList(SessionContext.LISTS);
            context.setMessage(context.get("Message.TooManyItems"));
            ResponsePage page = showListPropertiesResponse(context);

            if (page != null)
              return page;
          }
          catch (ClientException ignored) { }

        }
      }
    }

    String message = Util.getMessage(error);
    Properties subst = new Properties();
    subst.put("{_Title_}", error.getTitle());
    subst.put("{_Message_}", message);
    subst.put("{_StackTrace_}", ClientException.getStack(error));

    String urlName = context.getURL("{_HTMLDir_}/forms/Error.html");
    StringBuffer contents = null;

    try
    {
      contents = context.loadTemplateFile(urlName, subst);
    }
    catch (IOException xcpt)
    {
      context.log("Template file " + urlName + " not found: " + Util.toString(xcpt), true);
      contents = new StringBuffer(512);
      contents.append("<h2");
      if (context.getLocale().getLanguage().equals("iw"))
      {
      	contents.append(" dir=\"RTL\"");
      }
      contents.append(">").append(error.getTitle())
              .append("</h2>\r\n").append(message);
    }

    return new ResponsePage(contents);
  }

  /**
  * Creates the page that is displayed after a successful logon.
  * If there are worklists and the the display of work items is enabled, the
  * first worklist will be shown. Otherwise, if there are template lists and
  * the display of templates is enabled, the first template list will be shown.
  * Otherwise, if there are instance lists and the display of process instances
  * is enabled, the first instance list will be shown. If nothing is enabled that
  * is available, an exception will be thrown.
  * @param context The request context.
  * @return The response page.
  * @exception ClientException if an error occurred
  */
  public ResponsePage logonResponse(RequestContext context) throws ClientException
  {
    ResponsePage result = mostRecentList(context, false);

    if (result == null || context.getKindOfList() == SessionContext.LISTS) // nothing available that is enabled
      throw new ClientException(context, context.get("Error.NoLists"));

    return result;
  }

  /**
  * Creates the page that is shown after logoff.
  * This is the page specified by the <code>PageAfterLogoff</code> setting in
  * the <a href="../../../../../../properties.html">WebClient.properties</a>
  * configuration file.
  * @param context The request context.
  * @return The response page.
  */
  public ResponsePage logoffResponse(RequestContext context)
  {
    return new ResponsePage(context.getRedirect(pageAfterLogoff));
  }

  /** @doc
  * This method has an empty implementation since the Lists - Monitor
  * requires {@link JSPViewer}.
  * @param context The request context.
  * @return <code>null</code>.
  * @exception ClientException if an error occurred
  */
  public ResponsePage queryListsResponse(RequestContext context)
                      throws ClientException
  {
    return null;
  }

  /** @doc
  * This method has an empty implementation since the Lists - Monitor
  * requires {@link JSPViewer}.
  * @param context The request context.
  * @return <code>null</code>.
  * @exception ClientException if an error occurred
  */
  public ResponsePage showListPropertiesResponse(RequestContext context)
                      throws ClientException
  {
    return null;
  }

  /**
  * @doc
  */
  public ResponsePage setListPropertiesResponse(RequestContext context)
                      throws ClientException
  {
    String command = context.getCommand("createList");

    return new ResponsePage(context.getRedirect(command));
  }

  /**
  * Creates the page that is displayed when a list has been deleted.
  * This page calls {@link BuiltinHandler#queryLists BuiltinHandler.queryLists()}
  * to show the list of lists again.
  * @param context The request context.
  * @return The response page.
  * @exception ClientException if an error occurred
  */
  public ResponsePage deleteListResponse(RequestContext context)
                      throws ClientException
  {
    return mostRecentList(context, !autoRefresh);
  }

  /**
  * @doc
  */
  public ResponsePage createListResponse(RequestContext context)
                      throws ClientException
  {
    return null;
  }

  /**
  * @doc
  */
  public ResponsePage createPersistentListResponse(RequestContext context)
                      throws ClientException
  {
    PersistentList list = context.getWorkList();

    try
    {
      if (list != null)
      {
        context.setWorkListOid(list.persistentOid());
        return mostRecentList(context, false);
      }

      list = context.getInstanceList();

      if (list != null)
      {
        context.setInstanceListOid(list.persistentOid());
        return mostRecentList(context, false);
      }

      list = context.getTemplateList();

      if (list != null)
      {
        context.setTemplateListOid(list.persistentOid());
        return mostRecentList(context, false);
      }
      
      list = context.getActivityInstanceList();
      if (list != null)
      {
        context.setActivityInstanceListOid(list.persistentOid());
        return mostRecentList(context, false);
      }
    }
    catch (FmcException xcpt)
    {
      throw new ClientException(context, xcpt);
    }

    return null;
  }

  /**
  * Creates the page that displays a process template list.
  * This page is dynamically created and contains a table with a row
  * for each template on the list. This table also contains buttons that
  * allow to invoke other commands, such as displaying a template's properties.
  * @param context The request context.
  * @return The response page or <code>null</code> if template lists are
  *         disabled
  * @exception ClientException if an error occurred
  */
  public ResponsePage queryProcessTemplatesResponse(RequestContext context)
                      throws ClientException
  {
    if (!enableTemplatelists)
      return null;

    try
    {
      StringBuffer answer = new StringBuffer(10000);
      ProcessTemplateList templateList = context.getTemplateList();

      answer.append("<html>\r\n<head><title>");
      answer.append(context.get("Viewer.TemplateListTitle", templateList.name()));
      answer.append("</title></head>\r\n");
      answer.append("<body background=\"")
            .append(context.getURL("{_HTMLDir_}/images/background.gif"))
            .append("\" bgcolor=\"#C6C6C6\">\r\n");
      answer.append(context.openForm(Command.PTL_QUERY_TEMPLATES.getCommand(),
                                     templateList.persistentOid()));

      Vector navigatorEntries = createNavigatorEntries(context, true);
      String user = context.getUserID() + "@" + context.getSystemGroup();
      answer.append(createNavigatorComboBox(context, navigatorEntries, user));

      answer.append("<h2 align=\"left\">")
            .append(context.get("Viewer.TemplateListHeader",
                                new String[] { templateList.name(),
                                  String.valueOf(context.getTemplates().length) }));
      answer.append("</h2>\r\n");

      answer.append("<table border=\"0\" cellspacing=\"0\" width=\"100%\">");
      answer.append("<tr>\r\n");
      answer.append(" <th width=\"5%\"><b><i>")
            .append(context.get("Viewer.CommandHeader")).append("</i></b></th>\r\n");
      answer.append(" <th width=\"20%\"><b>")
            .append(context.get("Viewer.NameHeader")).append("</b></th>\r\n");
      answer.append(" <th width=\"5%\"></th>\r\n");
      answer.append(" <th width=\"40%\"><b>")
            .append(context.get("Viewer.DescriptionHeader")).append("</b></th>\r\n");
      answer.append("</tr>\r\n");
      answer.append("<tr><td width=\"100%\" colspan=\"4\"><hr></td></tr>\r\n");

      String bgcolors[] = { "white", "#CCCCFF" };

      for (int i = 0; i < context.getTemplates().length; ++i)
      {
        ProcessTemplate template = context.getTemplates()[i];
        String oid = template.persistentOid();
        String name = template.name();
        Command[] cmds = Command.getActions(template);

        answer.append("<tr bgcolor=\"")
              .append(bgcolors[i % bgcolors.length]).append("\">\r\n");
        answer.append("<td nowrap>");

        if (cmds.length == 0)
          answer.append("&nbsp;");
        else for (int j = 0; j < cmds.length; ++j)
          answer.append(cmds[j].getTriggerTag(context, oid, name));

        answer.append("</td>\r\n");

        answer.append("<td><small>").append(name)
              .append("</small></td>\r\n");
        answer.append("<td>")
              .append(Command.PT_SHOW_PROPERTIES.getTriggerTag(context, oid, name));
        answer.append("</td>\r\n");
        answer.append("<td nowrap><small>")
              .append(context.null2Empty(template.description()))
              .append("</small>");
        answer.append("</td>\r\n</tr>\r\n");
      }

      answer.append("<tr><td width=\"100%\" colspan=\"4\"><hr></td></tr>\r\n");
      answer.append("</table>");

      answer.append("<input type=\"submit\" value=\"");
      answer.append(context.get("Command.Refresh"))
            .append("\" name=\"refresh\">");
      answer.append("</form></body></html>\r\n");

	  return new ResponsePage(answer);
      //return redirectResponse(context, null, Command.PTL_QUERY_TEMPLATES,
      //                        templateList.persistentOid(), null);
    }
    catch (FmcException xcpt)
    {
      throw new ClientException(context, xcpt);
    }
  }

  /**
  * Creates the page that displays process template properties.
  * This page will be loaded from <code>forms/TemplateProperties.html</code>
  * and allows the following substitution variables:
  * <table border="1" cellspacing="0" cellpadding="4">
  *   <tr>
  *     <th>Name</th>
  *     <th>Replacement</th>
  *   </tr>
  *   <tr>
  *     <td>{_HTMLDir_}</td>
  *     <td>The root URI of the servlet's web application, for example,
  *         <code>/MQWFClient</code>.</td>
  *   </tr>
  *   <tr>
  *     <td>{_MQWFServlet_}</td>
  *     <td>The servlet's name, for example,
  *         <code>/MQWFClient/servlet/Main</code>.</td>
  *   </tr>
  *   <tr>
  *     <td>{Name}</td>
  *     <td>The template name.</td>
  *   </tr>
  *   <tr>
  *     <td>{Description}</td>
  *     <td>The template description.</td>
  *   </tr>
  *   <tr>
  *     <td>{Category}</td>
  *     <td>The category of the template.</td>
  *   </tr>
  *   <tr>
  *     <td>{InputDataStructure}</td>
  *     <td>The name of the template's input data structure.</td>
  *   </tr>
  *   <tr>
  *     <td>{OutputDataStructure}</td>
  *     <td>The name of the template's output data structure.</td>
  *   </tr>
  *   <tr>
  *     <td>{ProcessAdmin}</td>
  *     <td>The process administrator.</td>
  *   </tr>
  *   <tr>
  *     <td>{Role}</td>
  *     <td>The process' role.</td>
  *   </tr>
  *   <tr>
  *     <td>{Organization}</td>
  *     <td>The process' organization.</td>
  *   </tr>
  *   <tr>
  *     <td>{Documentation}</td>
  *     <td>The template documentation.</td>
  *   </tr>
  *   <tr>
  *     <td>{_ActionArea_}</td>
  *     <td>A form containing the <b>Back</b> button.</td>
  *   </tr>
  * </table>
  * @param context The request context.
  * @return The response page.
  * @exception ClientException if an error occurred
  */
  public ResponsePage showTemplatePropertiesResponse(RequestContext context)
                      throws ClientException
  {
    ProcessTemplate template = context.getTemplate();
    Properties subst = new Properties();

    try
    {
      template.refresh();

      subst.put("{Name}", null2Empty(template.name()));
      subst.put("{Description}", null2Empty(template.description()));
      subst.put("{Category}", null2Empty(template.category()));
      subst.put("{InputDataStructure}", null2Empty(template.inContainerName()));
      subst.put("{OutputDataStructure}", null2Empty(template.outContainerName()));
      subst.put("{ProcessAdmin}", null2Empty(template.processAdmin()));
      subst.put("{Role}", null2Empty(template.roleName()));
      subst.put("{Organization}", null2Empty(template.organizationName()));
      subst.put("{Documentation}", null2Empty(template.documentation()));
    }
    catch (FmcException xcpt)
    {
      throw new ClientException(context, xcpt);
    }

    StringBuffer form = context.openForm(Command.PTL_QUERY_TEMPLATES.getCommand(),
                                         context.getTemplateListOid());
    form.append("<input type=\"hidden\" name=\"useCache\" value=\"true\">\r\n");
    form.append("<input type=\"submit\" name=\"Back\"");
    form.append("value=\"").append(context.get("Viewer.BackToTemplatelist"))
        .append("\">\r\n</form>\r\n");
    subst.put("{_ActionArea_}", form.toString());

    String urlname = context.getURL("{_HTMLDir_}/forms/TemplateProperties.html");
    ResponsePage page = null;

    try
    {
      StringBuffer answer = context.loadTemplateFile(urlname, subst);
      page = new ResponsePage(answer);
    }
    catch (IOException xcpt)
    {
      context.log("Template file " + urlname + " could not be read: " + Util.toString(xcpt), true);
      throw new ClientException(context, context.get("Error.HTMLTemplateNotFound", urlname));
    }

    return page;
  }

  /**
  * Creates the page displayed after a new process instance has been created.
  * This page calls {@link BuiltinHandler#queryProcessTemplates BuiltinHandler.queryProcessTemplates()}
  * to show the template list again.
  * @param context The request context.
  * @return The response page.
  * @exception ClientException if an error occurred
  */
  public ResponsePage createInstanceResponse(RequestContext context)
                      throws ClientException
  {
    return mostRecentList(context, !autoRefresh);
  }

  /**
  * Creates the page displayed after a new process instance has been created
  * and started.
  * <p>If the process instance was not yet started because it
  * <a href="BuiltinHandler.html#data_reqd">required input data</a>, an HTML
  * template will be loaded from <code>processes/<i>TemplateName</i>.html</code>
  * which may contain the following substitution variables:
  * <table border="1" cellspacing="0" cellpadding="4">
  *   <tr>
  *     <th>Name</th>
  *     <th>Replacement</th>
  *   </tr>
  *   <tr>
  *     <td>{_HTMLDir_}</td>
  *     <td>The root URI of the servlet's web application, for example,
  *         <code>/MQWFClient</code>.</td>
  *   </tr>
  *   <tr>
  *     <td>{_MQWFServlet_}</td>
  *     <td>The servlet's name, for example,
  *         <code>/MQWFClient/servlet/Main</code>.</td>
  *   </tr>
  *   <tr>
  *     <td>{_ActionArea_}</td>
  *     <td>A form containing a <b>Submit</b> button which calls
  *         <code>createAndStartInstance</code> again with the
  *         <a href="BuiltinHandler.html#data_avail">submitData</a>
  *         request parameter set and with the input data entered, and
  *         a <b>Cancel</b> button which calls <code>queryProcessTemplates</code>
  *         to go back to the template list.</td>
  *   </tr>
  * </table>
  * If the HTML template file cannot be found, and
  * <a href="../../../../../../properties.html#generate">GenerateForms</a>
  * is set to true, a form is created from the process' input container.
  * Submitting this form will start the instance with the data entered in
  * the generated form.</p>
  * <p>If the process instance was started from a template list, this
  * template list will be displayed next.</p>
  * <p>If the process instance was started by the web user, the page denoted
  * by the <code>SuccessLocation</code> request parameter will be sent as
  * response. If <code>SuccessLocation</code> is not set, the default form
  * <code>forms/ProcessStarted.html</code> will be used. If additionally the
  * <code>UseRedirect</code> request parameter is set to <code>true</code>, the
  * request will be redirected to that page. If <code>UseRedirect</code> is
  * <code>false</code> or not set, the page
  * is loaded as an HTML template which may contain the following substitution
  * variables:
  * <table border="1" cellspacing="0" cellpadding="4">
  *   <tr>
  *     <th>Name</th>
  *     <th>Replacement</th>
  *   </tr>
  *   <tr>
  *     <td>{_HTMLDir_}</td>
  *     <td>The root URI of the servlet's web application, for example,
  *         <code>/MQWFClient</code>.</td>
  *   </tr>
  *   <tr>
  *     <td>{_MQWFServlet_}</td>
  *     <td>The servlet's name, for example,
  *         <code>/MQWFClient/servlet/Main</code>.</td>
  *   </tr>
  *   <tr>
  *     <td>{_TemplateName_}</td>
  *     <td>The name of the process template.</td>
  *   </tr>
  *   <tr>
  *     <td>{_InstanceName_}</td>
  *     <td>The name of the process instance that has been started.</td>
  *   </tr>
  * </table>
  * </p>
  * @param context The request context.
  * @return The response page.
  * @exception ClientException if an error occurred
  */
  public ResponsePage createAndStartInstanceResponse(RequestContext context)
                      throws ClientException
  {
    ResponsePage page = null;

    try
    {
      ProcessInstance instance = context.getInstance();
      String templateName = context.getTemplate().name();
      String templatelistOid = context.getTemplateListOid();
      String oid = context.getRequestParameter("id", null);

      if (instance == null)
      {
        // not yet started because input data required
        String urlname = context.getURL("{_HTMLDir_}/processes/" + templateName + ".html");

        StringBuffer actionArea = new StringBuffer(1024);
        actionArea.append("<input type=\"hidden\" name=\"command\"")
                  .append(" value=\"")
                  .append(Command.PT_CREATE_AND_START_INSTANCE.getCommand())
                  .append("\">\r\n");
        actionArea.append("<input type=\"hidden\" name=\"id\"")
                  .append(" value=\"").append(oid)
                  .append("\">\r\n");
        actionArea.append("<input type=\"submit\" name=\"submitData\"")
                  .append(" value=\"")
                  .append(context.get("Command.StartInstance")).append("\">\r\n");
        actionArea.append("<input type=\"button\" name=\"cancel\"")
                  .append(" value=\"").append(context.get("Command.Cancel"))
                  .append("\" onClick=\"javascript:location.replace('")
                  .append(Command.PTL_QUERY_TEMPLATES.withParameter("&useCache=true")
                                 .getURL(context, templatelistOid, null, false))
                  .append("')\">\r\n");

        Properties subst = new Properties();
        subst.put("{_ActionArea_}", actionArea.toString());

        try
        {
          StringBuffer answer = context.loadTemplateFile(urlname, subst);
          page = new ResponsePage(answer);
        }
        catch (IOException xcpt)
        {
          if (!generateForms)
            throw new ClientException(context, context.get("Error.HTMLTemplateNotFound", urlname));

          context.log("Template file " + urlname + " could not be read: " + Util.toString(xcpt), false);

          ReadWriteContainer inData = context.getTemplate().initialInContainer();

          StringBuffer answer = context.openPage(context.get("Viewer.CreateInstanceTitle"), 2048);
          answer.append("<body>\r\n");

          // Create a table that holds all input fields
          answer.append("<h2>")
                .append(context.get("Viewer.CreateInstanceHeader", templateName))
                .append("</h2>\r\n");
          answer.append("<br>\r\n");

          answer.append(context.generateForm(Command.PT_CREATE_AND_START_INSTANCE, oid,
                                             Command.PTL_QUERY_TEMPLATES.withText("Command.Cancel")
                                                    .withParameter("&useCache=true"),
                                             templatelistOid, null, inData, templateName, null));

          answer.append("</body></html>");

          return new ResponsePage(answer);
        }
      }
      else
      {
        // instance != null - process successfully started
        String urlname = null;
        String useRedirect = null;

        if (oid != null)
        {
          // process was started from template list
          urlname = Command.PTL_QUERY_TEMPLATES.withParameter("&useCache=true")
                           .getURL(context, templatelistOid, null, false);
          useRedirect = Boolean.TRUE.toString();
        }
        else
        {
          // process was started by the webuser
          urlname = context.getRequestParameter("SuccessLocation",
                                                "{_HTMLDir_}/forms/ProcessStarted.html");
          urlname = context.getURL(urlname);
          useRedirect = context.getRequestParameter("UseRedirect", Boolean.FALSE.toString());
        }

        if (Util.booleanValue(useRedirect, false))
        {
          page = new ResponsePage(context.getRedirect(urlname));
        }
        else
        {
          Properties subst = new Properties();
          subst.put("{_InstanceName_}", instance.name());
          subst.put("{_TemplateName_}", templateName);

          try
          {
            StringBuffer answer = context.loadTemplateFile(urlname, subst);
            page = new ResponsePage(answer);
          }
          catch (IOException xcpt)
          {
            context.log("Template file " + urlname + " could not be read: " + Util.toString(xcpt), true);
            throw new ClientException(context, context.get("Error.HTMLTemplateNotFound", urlname));
          }
        }
      }
    }
    catch (FmcException xcpt)
    {
      throw new ClientException(context, xcpt);
    }

    return page;
  }

  /**
  * Creates the page displayed after a process template has been deleted.
  * This page calls {@link BuiltinHandler#queryProcessTemplates BuiltinHandler.queryProcessTemplates()}
  * to show the template list again.
  * @param context the request context
  * @return The response page.
  * @exception ClientException if an error occurred
  */
  public ResponsePage deleteTemplateResponse(RequestContext context)
                      throws ClientException
  {
    return mostRecentList(context, !autoRefresh);
  }

  /**
  * Creates the page that displays a process instance list.
  * This page is dynamically created and contains a table with a row
  * for each instance on the list. This table also contains buttons that
  * allow to invoke other commands, such as displaying an instance's properties.
  * @param context The request context.
  * @return The response page.
  * @exception ClientException if an error occurred
  */
  public ResponsePage queryProcessInstancesResponse(RequestContext context)
                      throws ClientException
  {
    if (!enableInstancelists)
      return null;

    try
    {
      StringBuffer answer = new StringBuffer(10000);
      ProcessInstanceList instanceList = context.getInstanceList();

      answer.append("<html>\r\n<head><title>");
      answer.append(context.get("Viewer.InstanceListTitle", instanceList.name()));
      answer.append("</title></head>\r\n");
      answer.append("<body background=\"")
            .append(context.getURL("{_HTMLDir_}/images/background.gif"))
            .append("\" bgcolor=\"#C6C6C6\">\r\n");
      answer.append(context.openForm(Command.PIL_QUERY_INSTANCES.getCommand(),
                                     instanceList.persistentOid()));

      Vector navigatorEntries = createNavigatorEntries(context, true);
      String user = context.getUserID() + "@" + context.getSystemGroup();
      answer.append(createNavigatorComboBox(context, navigatorEntries, user));

      answer.append("<h2 align=\"left\">")
            .append(context.get("Viewer.InstanceListHeader",
                                new String[] { instanceList.name(),
                                  String.valueOf(context.getInstances().length) }));
      answer.append("</h2>\r\n");

      answer.append("<table border=\"0\" cellspacing=\"0\" width=\"100%\">");
      answer.append("<tr>\r\n");
      answer.append(" <th width=\"10%\"><b><i>")
            .append(context.get("Viewer.CommandHeader")).append("</i></b></th>\r\n");
      answer.append(" <th width=\"2%\"><b>")
            .append(context.get("Viewer.StateHeader")).append("</b></th>\r\n");
      answer.append(" <th width=\"25%\"><b>")
            .append(context.get("Viewer.NameHeader")).append("</b></th>\r\n");
      answer.append(" <th width=\"3%\"></th>\r\n"); // Properties button
      answer.append(" <th width=\"40%\"><b>")
            .append(context.get("Viewer.DescriptionHeader")).append("</b></th>\r\n");
      answer.append(" <th width=\"20%\"><b>")
            .append(context.get("Viewer.LastModifiedHeader")).append("</b></th>\r\n");
      answer.append("</tr>\r\n");
      answer.append("<tr><td width=\"100%\" colspan=\"6\"><hr></td></tr>\r\n");

      String bgcolors[] = { "white", "#CCCCFF" };

      for (int i = 0; i < context.getInstances().length; ++i)
      {
        try
        {
          ProcessInstance instance = context.getInstances()[i];
          String oid = instance.persistentOid();
          String name = instance.name();
          Command[] cmds = Command.getActions(instance);

          answer.append("<tr bgcolor=\"")
                .append(bgcolors[i % bgcolors.length]).append("\">\r\n");
          answer.append("<td nowrap>");

          if (cmds.length == 0)
            answer.append("&nbsp;");
          else for (int j = 0; j < cmds.length; ++j)
            answer.append(cmds[j].getTriggerTag(context, oid, name));

          answer.append("</td>\r\n");

          answer.append("<td><img src=\"")
                .append(context.getURL("{_HTMLDir_}/images/state/" +
                        context.getPIImageForState(instance.state())))
                .append("\" \r\n");
          answer.append("alt=\"")
                .append(context.getPIStringForState(instance.state()))
                .append("\"></td>\r\n");
          answer.append("<td nowrap><small>").append(name)
                .append("</small></td>\r\n");
          answer.append("<td>")
                .append(Command.PI_SHOW_PROPERTIES.getTriggerTag(context, oid, name))
                .append("</td>\r\n");
          answer.append("<td nowrap><small>&nbsp;")
                .append(context.null2Empty(instance.description()))
                .append("</small></td>\r\n");
          answer.append("<td nowrap><small>")
                .append(context.toString(instance.lastModificationTime()))
                .append("</small></td>\r\n");
          answer.append("</tr>\r\n");
        }
        catch (FmcException xcpt)
        {
          if (xcpt.rc != xcpt.FMC_ERROR_DOES_NOT_EXIST)
            throw xcpt;
        }
      }

      answer.append("<tr><td width=\"100%\" colspan=\"6\"><hr></td></tr>\r\n");
      answer.append("</table>");

      answer.append("<input type=\"submit\" value=\"");
      answer.append(context.get("Command.Refresh"));
      answer.append("\" name=\"Refresh\">");
      answer.append("</form></body></html>\r\n");

	  return new ResponsePage(answer);
      //return redirectResponse(context, null, Command.PIL_QUERY_INSTANCES,
      //                        instanceList.persistentOid(), null);
    }
    catch (FmcException xcpt)
    {
      throw new ClientException(context, xcpt);
    }
  }

  /**
  * Creates the page that displays process instance properties.
  * This page will be loaded from <code>forms/InstanceProperties.html</code>
  * and allows the following substitution variables:
  * <table border="1" cellspacing="0" cellpadding="4">
  *   <tr>
  *     <th>Name</th>
  *     <th>Replacement</th>
  *   </tr>
  *   <tr>
  *     <td>{_HTMLDir_}</td>
  *     <td>The root URI of the servlet's web application, for example,
  *         <code>/MQWFClient</code>.</td>
  *   </tr>
  *   <tr>
  *     <td>{_MQWFServlet_}</td>
  *     <td>The servlet's name, for example,
  *         <code>/MQWFClient/servlet/Main</code>.</td>
  *   </tr>
  *   <tr>
  *     <td>{Name}</td>
  *     <td>The instance name.</td>
  *   </tr>
  *   <tr>
  *     <td>{Description}</td>
  *     <td>The instance description.</td>
  *   </tr>
  *   <tr>
  *     <td>{Category}</td>
  *     <td>The category of the instance.</td>
  *   </tr>
  *   <tr>
  *     <td>{State}</td>
  *     <td>The process state.</td>
  *   </tr>
  *   <tr>
  *     <td>{ParentName}</td>
  *     <td>The parent process name.</td>
  *   </tr>
  *   <tr>
  *     <td>{TopLevelName}</td>
  *     <td>The top level process name.</td>
  *   </tr>
  *   <tr>
  *     <td>{InputDataStructure}</td>
  *     <td>The name of the template's input data structure.</td>
  *   </tr>
  *   <tr>
  *     <td>{OutputDataStructure}</td>
  *     <td>The name of the template's output data structure.</td>
  *   </tr>
  *   <tr>
  *     <td>{Starter}</td>
  *     <td>The process starter.</td>
  *   </tr>
  *   <tr>
  *     <td>{ProcessAdmin}</td>
  *     <td>The process administrator.</td>
  *   </tr>
  *   <tr>
  *     <td>{Role}</td>
  *     <td>The process' role.</td>
  *   </tr>
  *   <tr>
  *     <td>{Organization}</td>
  *     <td>The process' organization.</td>
  *   </tr>
  *   <tr>
  *     <td>{Documentation}</td>
  *     <td>The process documentation.</td>
  *   </tr>
  *   <tr>
  *     <td>{Received}</td>
  *     <td>The process start time.</td>
  *   </tr>
  *   <tr>
  *     <td>{Notification}</td>
  *     <td>The process notification time.</td>
  *   </tr>
  *   <tr>
  *     <td>{Finished}</td>
  *     <td>The process end time.</td>
  *   </tr>
  *   <tr>
  *     <td>{_ActionArea_}</td>
  *     <td>A form containing the <b>Back</b> button.</td>
  *   </tr>
  * </table>
  * @param context The request context.
  * @return The response page.
  * @exception ClientException if an error occurred
  */
  public ResponsePage showInstancePropertiesResponse(RequestContext context)
                      throws ClientException
  {
    ProcessInstance instance = context.getInstance();
    Properties subst = new Properties();

    try
    {
      instance.refresh();

      subst.put("{Name}", null2Empty(instance.name()));
      subst.put("{Description}", null2Empty(instance.description()));
      subst.put("{Category}", null2Empty(instance.category()));
      subst.put("{State}", null2Empty(context.getPIStringForState(instance.state())));
      subst.put("{ParentName}", null2Empty(instance.parentName()));
      subst.put("{TopLevelName}", null2Empty(instance.topLevelName()));
      subst.put("{InputDataStructure}", null2Empty(instance.inContainerName()));
      subst.put("{OutputDataStructure}", null2Empty(instance.outContainerName()));
      subst.put("{Starter}", null2Empty(instance.starter()));
      subst.put("{ProcessAdmin}", null2Empty(instance.processAdmin()));
      subst.put("{Role}", null2Empty(instance.roleName()));
      subst.put("{Organization}", null2Empty(instance.organizationName()));
      subst.put("{Documentation}", null2Empty(instance.documentation()));
      subst.put("{Received}", context.toString(instance.startTime()));
      subst.put("{Notification}", context.toString(instance.notificationTime()));
      subst.put("{Finished}", context.toString(instance.endTime()));
    }
    catch(FmcException xcpt)
    {
      throw new ClientException(context, xcpt);
    }

    StringBuffer form = context.openForm(Command.PIL_QUERY_INSTANCES.getCommand(),
                                         context.getInstanceListOid());
    form.append("<input type=\"hidden\" name=\"useCache\" value=\"true\">\r\n");
    form.append("<input type=\"submit\" name=\"Back\" ");
    form.append("value=\"").append(context.get("Viewer.BackToInstancelist")).append("\">\r\n");
    form.append("</form>\r\n");
    subst.put("{_ActionArea_}", form.toString());

    String urlname = context.getURL("{_HTMLDir_}/forms/InstanceProperties.html");
    ResponsePage page = null;

    try
    {
      StringBuffer answer = context.loadTemplateFile(urlname, subst);
      page = new ResponsePage(answer);
    }
    catch (IOException xcpt)
    {
      context.log("Template file " + urlname + " could not be read: " + Util.toString(xcpt), true);
      throw new ClientException(context, context.get("Error.HTMLTemplateNotFound", urlname));
    }

    return page;
  }

  /**
  * Creates the page that will be displayed after a process instance has
  * been started.
  * If the process instance was actually started, the instance list will
  * be shown again. If the process instance was not yet started because it
  * required input data, an HTML template will be loaded from
  * <code>processes/<i>TemplateName</i>.html</code> which may
  * contain the following substitution variables:
  * <table border="1" cellspacing="0" cellpadding="4">
  *   <tr>
  *     <th>Name</th>
  *     <th>Replacement</th>
  *   </tr>
  *   <tr>
  *     <td>{_HTMLDir_}</td>
  *     <td>The root URI of the servlet's web application, for example,
  *         <code>/MQWFClient</code>.</td>
  *   </tr>
  *   <tr>
  *     <td>{_MQWFServlet_}</td>
  *     <td>The servlet's name, for example,
  *         <code>/MQWFClient/servlet/Main</code>.</td>
  *   </tr>
  *   <tr>
  *     <td>{_ActionArea_}</td>
  *     <td>A form containing a <b>Submit</b> button which calls
  *         <code>startInstance</code> with the input data entered and
  *         a <b>Cancel</b> button which calls <code>queryProcessInstances</code>
  *         to go back to the instance list.</td>
  *   </tr>
  * </table>
  * If the HTML template file cannot be found, and
  * <a href="../../../../../../properties.html#generate">GenerateForms</a>
  * is set to true, a form is created from the process' input container.
  * Submitting this form will start the instance with the data entered in
  * the generated form.
  * @param context The request context.
  * @return The response page.
  * @exception ClientException if an error occurred
  */
  public ResponsePage startInstanceResponse(RequestContext context)
                      throws ClientException
  {
    ProcessInstance instance = context.getInstance();
    ResponsePage page = null;

    try
    {
      if (instance.state().value() !=
          com.ibm.workflow.api.ProcessInstancePackage.ExecutionState._READY)
      {
        page = mostRecentList(context, !autoRefresh);
      }
      else
      {
        // instance.state() == READY - not yet started because input data required
        String templateName = instance.processTemplateName();
        String urlname = context.getURL("{_HTMLDir_}/processes/" + templateName + ".html");
        String instancelistOid = context.getInstanceListOid();
        String oid = instance.persistentOid();

        StringBuffer actionArea = new StringBuffer(1024);
        actionArea.append("<input type=\"hidden\" name=\"command\"")
                  .append(" value=\"")
                  .append(Command.PI_START.getCommand())
                  .append("\">\r\n");
        actionArea.append("<input type=\"hidden\" name=\"id\"")
                  .append(" value=\"").append(oid)
                  .append("\">\r\n");
        actionArea.append("<input type=\"submit\" name=\"submitData\"")
                  .append(" value=\"")
                  .append(context.get("Command.StartInstance"))
                  .append("\">\r\n");
        actionArea.append("<input type=\"button\" name=\"cancel\"")
                  .append(" value=\"").append(context.get("Command.Cancel"))
                  .append("\" onClick=\"javascript:location.replace('")
                  .append(Command.PIL_QUERY_INSTANCES.withParameter("&useCache=true")
                                 .getURL(context, instancelistOid, null, false))
                  .append("')\">\r\n");

        Properties subst = new Properties();
        subst.put("{_ActionArea_}", actionArea.toString());

        try
        {
          StringBuffer answer = context.loadTemplateFile(urlname, subst);
          page = new ResponsePage(answer);
        }
        catch (IOException xcpt)
        {
          if (!generateForms)
            throw new ClientException(context, context.get("Error.HTMLTemplateNotFound", urlname));

          context.log("Template file " + urlname + " could not be read: " + Util.toString(xcpt), false);

          ReadWriteContainer inData = instance.inContainer();

          StringBuffer answer = context.openPage(context.get("Viewer.StartInstanceTitle"), 2048);
          answer.append("<body>\r\n");

          // Create a table that holds all input fields
          answer.append("<h2>")
                .append(context.get("Viewer.StartInstanceHeader", instance.name()))
                .append("</h2>\r\n");
          answer.append("<br><br>\r\n");

          answer.append(context.generateForm(Command.PI_START, oid,
                                             Command.PIL_QUERY_INSTANCES.withText("Command.Cancel")
                                                    .withParameter("&useCache=true"),
                                             instancelistOid, null, inData, instance.name(), null));

          answer.append("</body></html>");

          return new ResponsePage(answer);
        }
      }
    }
    catch (FmcException xcpt)
    {
      throw new ClientException(context, xcpt);
    }

    return page;
  }

  /**
  * Creates the page displayed after a process instance has been deleted.
  * This page calls {@link BuiltinHandler#queryProcessInstances BuiltinHandler.queryProcessInstances()}
  * to show the instance list again.
  * @param context the request context
  * @return The response page.
  * @exception ClientException if an error occurred
  */
  public ResponsePage deleteInstanceResponse(RequestContext context)
                      throws ClientException
  {
    return mostRecentList(context, !autoRefresh);
  }

  /**
  * Creates the page displayed after a process instance has been suspended.
  * If the process instance was actually suspended, this method simply calls
  * the {@link #queryProcessInstancesResponse queryProcessInstancesResponse()}
  * method. If the instance was not yet suspended because the necessary
  * data has not yet been specified, it will be suspended first with the
  * <code>deep</code> parameter set to <code>true</code> and no 'until' date
  * specified.
  * @param context the request context
  * @return The response page.
  * @exception ClientException if an error occurred
  */
  public ResponsePage suspendInstanceResponse(RequestContext context)
                      throws ClientException
  {
    if (context.getRequestParameter("submitData", null) == null)
    {
      try
      {
        BuiltinHandler builtin = context.getBuiltinHandler();
        ProcessInstance instance = context.getInstance();
        builtin.suspendInstance(context, instance.persistentOid(), "true",
                                true, null, null, null, null, null);
      }
      catch (FmcException xcpt)
      {
        throw new ClientException(context, xcpt);
      }
    }

    return mostRecentList(context, !autoRefresh);
  }

  /**
  * Creates the page displayed after a process instance has been resumed.
  * This page calls {@link BuiltinHandler#queryProcessInstances BuiltinHandler.queryProcessInstances()}
  * to show the instance list again.
  * @param context the request context
  * @return The response page.
  * @exception ClientException if an error occurred
  */
  public ResponsePage resumeInstanceResponse(RequestContext context)
                      throws ClientException
  {
    return mostRecentList(context, !autoRefresh);
  }

  /**
  * Creates the page displayed after a process instance has been terminated.
  * This page calls {@link BuiltinHandler#queryProcessInstances BuiltinHandler.queryProcessInstances()}
  * to show the instance list again.
  * @param context the request context
  * @return The response page.
  * @exception ClientException if an error occurred
  */
  public ResponsePage terminateInstanceResponse(RequestContext context)
                      throws ClientException
  {
    return mostRecentList(context, !autoRefresh);
  }

  /**
  * Creates the page displayed after a process instance has been restarted.
  * This page calls {@link BuiltinHandler#queryProcessInstances BuiltinHandler.queryProcessInstances()}
  * @param context the request context
  * @return The response page.
  * @exception ClientException if an error occurred
  */
  public ResponsePage restartInstanceResponse(RequestContext context)
                      throws ClientException
  {
    return mostRecentList(context, !autoRefresh);
  }

  /**
  * This method has an empty implementation since activity properties
  * can only be accessed from within the process monitor which requires
  * {@link JSPViewer}.
  * @param context The request context.
  * @return <code>null</code>.
  * @exception ClientException if an error occurred
  */
  public ResponsePage showActivityPropertiesResponse(RequestContext context)
                      throws ClientException
  {
    return null;
  }

  /**
  * This method has an empty implementation since activity actions
  * can only be accessed from activity properties page which requires
  * {@link JSPViewer}.
  * @param context The request context.
  * @return <code>null</code>.
  * @exception ClientException if an error occurred
  */
  public ResponsePage forceRestartActivityResponse(RequestContext context)
                      throws ClientException
  {
    return null;
  }

  /**
  * This method has an empty implementation since activity actions
  * can only be accessed from activity properties page which requires
  * {@link JSPViewer}.
  * @param context The request context.
  * @return <code>null</code>.
  * @exception ClientException if an error occurred
  */
  public ResponsePage forceFinishActivityResponse(RequestContext context)
                      throws ClientException
  {
    return null;
  }

  /**
  * This method has an empty implementation since activity actions
  * can only be accessed from activity properties page which requires
  * {@link JSPViewer}.
  * @param context The request context.
  * @return <code>null</code>.
  * @exception ClientException if an error occurred
  */
  public ResponsePage terminateActivityResponse(RequestContext context)
                      throws ClientException
  {
    return null;
  }

  /**
  * This method has an empty implementation since connector properties
  * can only be accessed from within the process monitor which requires
  * {@link JSPViewer}.
  * @param context The request context.
  * @return <code>null</code>.
  * @exception ClientException if an error occurred
  */
  public ResponsePage showControlConnectorPropertiesResponse(RequestContext context)
                      throws ClientException
  {
    return null;
  }

  /**
  * Creates the page that displays a work list.
  * This page is dynamically created and contains a table with a row
  * for each item on the list. This table also contains buttons that
  * allow to invoke other commands, such as displaying an item's properties.
  * @param context The request context.
  * @return The response page.
  * @exception ClientException if an error occurred
  */
  public ResponsePage queryWorkItemsResponse(RequestContext context)
                      throws ClientException
  {
    if (!enableWorklists)
      return null;
    
    try
    {
      
      StringBuffer answer = new StringBuffer(10000);
      WorkList workList = context.getWorkList();
      answer.append("<html>").append("\r\n<head><title>");
      answer.append(context.get("Viewer.WorklistTitle", workList.name()));
      answer.append("</title></head>\r\n");
      answer.append("<body background=\"")
            .append(context.getURL("{_HTMLDir_}/images/background.gif"))
            .append("\" bgcolor=\"#C6C6C6\">\r\n");
      answer.append(context.openForm(Command.WL_QUERY_WORK_ITEMS.getCommand(),
                                     workList.persistentOid()));

      String user = context.getUserID();
      String fqUser = user + "@" + context.getSystemGroup();
      Vector navigatorEntries = createNavigatorEntries(context, true);

      answer.append(createNavigatorComboBox(context, navigatorEntries, fqUser));
      answer.append("<h2 align=\"left\">")
            .append(context.get("Viewer.WorklistHeader",
                                new String[] { workList.name(),
                                  String.valueOf(context.getWorkItems().length) }));
      answer.append("</h2>\r\n");

      answer.append("<table border=\"0\" cellspacing=\"0\" width=\"100%\">");

      answer.append("<tr>\r\n");
      answer.append(" <th width=\"5%\"><b><i>")
            .append(context.get("Viewer.CommandHeader")).append("</i></b></th>\r\n");
      answer.append(" <th width=\"1%\" align=\"center\"><b>")
            .append(context.get("Viewer.StateHeader")).append("</b></th>\r\n");
      answer.append(" <th width=\"20%\"><b>")
            .append(context.get("Viewer.NameHeader")).append("</b></th>\r\n");
      answer.append(" <th width=\"5%\"></th>\r\n"); // Properties button
      answer.append(" <th width=\"30%\"><b>")
            .append(context.get("Viewer.DescriptionHeader")).append("</b></th>\r\n");
      answer.append(" <th width=\"15%\"><b>")
            .append(context.get("Viewer.InstanceNameHeader")).append("</b></th>\r\n");
      answer.append(" <th width=\"9%\"><b>")
            .append(context.get("Viewer.OwnerHeader")).append("</b></th>\r\n");
      answer.append(" <th width=\"15%\"><b>")
            .append(context.get("Viewer.CreationHeader")).append("</b></th>\r\n");
      answer.append("</tr>\r\n");
      answer.append("<tr><td width=\"100%\" colspan=\"8\"><hr></td></tr>\r\n\r\n");

      String bgcolors[] = { "white", "#CCCCFF" };

      for (int i = 0; i < context.getWorkItems().length; i++)
      {
        try
        {
          WorkItem workItem = context.getWorkItems()[i];
          String oid = workItem.persistentOid();
          String name = workItem.name();
          Command[] cmds = Command.getActions(user, workItem);

          answer.append("<tr bgcolor=\"")
                .append(bgcolors[i % bgcolors.length]).append("\">\r\n");
          answer.append("<td nowrap>");

          if (cmds.length == 0)
            answer.append("&nbsp;");
          else for (int j = 0; j < cmds.length; ++j)
          {
            if ("transferItem".equals(cmds[j].getCommand()))
              answer.append(cmds[j].getTriggerTag(context, oid, user));
            else
              answer.append(cmds[j].getTriggerTag(context, oid, name));
          }

          answer.append("</td>");
          answer.append("<td>");
          answer.append("<img src=\"")
                .append(context.getURL("{_HTMLDir_}/images/state/" +
                        context.getWIImageForState(workItem.state())))
                .append("\" ").append("alt=\"")
                .append(context.getWIStringForState(workItem.state()))
                .append("\"></td>");
          answer.append("<td><small>").append(workItem.name())
                .append("</small></td>\r\n");
          answer.append("<td>")
                .append(Command.WI_SHOW_PROPERTIES.getTriggerTag(context, oid, name))
                .append("</td>\r\n");
          answer.append("<td nowrap><small>").append(context.null2Empty(workItem.description()))
                .append("</small></td>\r\n");
          answer.append("<td nowrap><small>").append(workItem.processInstanceName())
                .append("&nbsp;</small></td>\r\n");
          answer.append("<td><small>").append(workItem.owner())
                .append("&nbsp;</small></td>\r\n");
          answer.append("<td nowrap><small>")
                .append(context.toString(workItem.creationTime()))
                .append("</small></td>");
          answer.append("</tr>\r\n");
        }
        catch (FmcException xcpt)
        {
          if (xcpt.rc != xcpt.FMC_ERROR_DOES_NOT_EXIST)
            throw xcpt;
        }
      }

      answer.append("<tr><td width=\"100%\" colspan=\"8\"><hr></td></tr>\r\n");
      answer.append("</table>");

      answer.append("<input type=\"submit\" value=\"");
      answer.append(context.get("Command.Refresh")).append("\" name=\"Refresh\">");
      answer.append("</form></body></html>\r\n");

      return new ResponsePage(answer);
      //return redirectResponse(context, null, Command.WL_QUERY_WORK_ITEMS,
      //                        workList.persistentOid(), null);
      
    }
    catch (FmcException xcpt)
    {
      throw new ClientException(context, xcpt);
    }
  }

  /**
  * Creates the page that displays work item properties.
  * This page will be loaded from <code>forms/WorkItemProperties.html</code>
  * and allows the following substitution variables:
  * <table border="1" cellspacing="0" cellpadding="4">
  *   <tr>
  *     <th>Name</th>
  *     <th>Replacement</th>
  *   </tr>
  *   <tr>
  *     <td>{_HTMLDir_}</td>
  *     <td>The root URI of the servlet's web application, for example,
  *         <code>/MQWFClient</code>.</td>
  *   </tr>
  *   <tr>
  *     <td>{_MQWFServlet_}</td>
  *     <td>The servlet's name, for example,
  *         <code>/MQWFClient/servlet/Main</code>.</td>
  *   </tr>
  *   <tr>
  *     <td>{Name}</td>
  *     <td>The item name.</td>
  *   </tr>
  *   <tr>
  *     <td>{Description}</td>
  *     <td>The item description.</td>
  *   </tr>
  *   <tr>
  *     <td>{State}</td>
  *     <td>The item state.</td>
  *   </tr>
  *   <tr>
  *     <td>{Type}</td>
  *     <td>The item type.</td>
  *   </tr>
  *   <tr>
  *     <td>{ProgramActivity}</td>
  *     <td>The item implementation.</td>
  *   </tr>
  *   <tr>
  *     <td>{ReceivedAs}</td>
  *     <td>Why the item was assigned.</td>
  *   </tr>
  *   <tr>
  *     <td>{ProcessName}</td>
  *     <td>The parent process name.</td>
  *   </tr>
  *   <tr>
  *     <td>{Category}</td>
  *     <td>The item category.</td>
  *   </tr>
  *   <tr>
  *     <td>{Start}</td>
  *     <td>The activity's start mode.</td>
  *   </tr>
  *   <tr>
  *     <td>{StartCondition}</td>
  *     <td>The activity's start condition.</td>
  *   </tr>
  *   <tr>
  *     <td>{Exit}</td>
  *     <td>The activity's exit mode.</td>
  *   </tr>
  *   <tr>
  *     <td>{ExitCondition}</td>
  *     <td>The activity's exit condition.</td>
  *   </tr>
  *   <tr>
  *     <td>{OnTheWorkListOf}</td>
  *     <td>The persons who have a work item assigned.</td>
  *   </tr>
  *   <tr>
  *     <td>{ProcessAdmin}</td>
  *     <td>The process administrator.</td>
  *   </tr>
  *   <tr>
  *     <td>{Priority}</td>
  *     <td>The item's priority.</td>
  *   </tr>
  *   <tr>
  *     <td>{InputDataStructure}</td>
  *     <td>The name of the template's input data structure.</td>
  *   </tr>
  *   <tr>
  *     <td>{OutputDataStructure}</td>
  *     <td>The name of the template's output data structure.</td>
  *   </tr>
  *   <tr>
  *     <td>{Documentation}</td>
  *     <td>The process documentation.</td>
  *   </tr>
  *   <tr>
  *     <td>{Received}</td>
  *     <td>The item's start time.</td>
  *   </tr>
  *   <tr>
  *     <td>{FirstNotification}</td>
  *     <td>The item's first notification time.</td>
  *   </tr>
  *   <tr>
  *     <td>{SecondNotification}</td>
  *     <td>The item's second notification time.</td>
  *   </tr>
  *   <tr>
  *     <td>{Finished}</td>
  *     <td>The item's end time.</td>
  *   </tr>
  *   <tr>
  *     <td>{_ActionArea_}</td>
  *     <td>A form containing the <b>Back</b> button.</td>
  *   </tr>
  * </table>
  * @param context The request context.
  * @return The response page.
  * @exception ClientException if an error occurred
  */
  public ResponsePage showWorkItemPropertiesResponse(RequestContext context)
                      throws ClientException
  {
    WorkItem workItem = context.getWorkItem();
    Properties subst = new Properties();

    try
    {
      workItem.refresh();

      subst.put("{Name}", null2Empty(workItem.name()));
      subst.put("{Description}", null2Empty(workItem.description()));
      subst.put("{State}", context.getWIStringForState(workItem.state()));
      subst.put("{Type}", context.toString(workItem.kind()));
      subst.put("{ProgramActivity}", null2Empty(workItem.implementation()));
      subst.put("{ReceivedAs}", context.toString(workItem.receivedAs()));
      subst.put("{ProcessName}", null2Empty(workItem.processInstanceName()));
      subst.put("{Category}", null2Empty(workItem.category()));
      subst.put("{Start}", workItem.manualStartMode () ?
                              context.get("Viewer.ModeManual") :
                              context.get("Viewer.ModeAutomatic"));
      subst.put("{StartCondition}", null2Empty(workItem.startCondition()));
      subst.put("{Exit}", workItem.manualExitMode () ?
                              context.get("Viewer.ModeManual") :
                              context.get("Viewer.ModeAutomatic"));
      subst.put("{ExitCondition}", null2Empty(workItem.exitCondition()));

      String[] staff = workItem.staff();
      StringBuffer staffBuffer = new StringBuffer(1000);

      for (int i = 0; i < staff.length; ++i)
        staffBuffer.append (staff[i] + " ");

      subst.put("{OnTheWorklistOf}", null2Empty(staffBuffer.toString()));
      subst.put("{ProcessAdmin}", null2Empty(workItem.processAdmin()));
      subst.put("{Priority}", String.valueOf(workItem.priority()));
      subst.put("{InputDataStructure}", null2Empty(workItem.inContainerName()));
      subst.put("{OutputDataStructure}", null2Empty(workItem.outContainerName()));
      subst.put("{Documentation}", null2Empty(workItem.documentation()));
      subst.put("{Received}", context.toString(workItem.startTime()));
      subst.put("{FirstNotification}", context.toString(workItem.firstNotificationTime()));
      subst.put("{SecondNotification}", context.toString(workItem.secondNotificationTime()));
      subst.put("{Finished}", context.toString(workItem.endTime()));
    }
    catch(FmcException xcpt)
    {
      throw new ClientException(context, xcpt);
    }

    StringBuffer form = context.openForm(Command.WL_QUERY_WORK_ITEMS.getCommand(),
                                         context.getWorkListOid());
    form.append("<input type=\"hidden\" name=\"useCache\" value=\"true\">\r\n");
    form.append("<input type=\"submit\" name=\"Back\" ");
    form.append("value=\"" + context.get("Viewer.BackToWorklist") + "\">\r\n");
    form.append("</form>\r\n");
    subst.put("{_ActionArea_}", form.toString());

    String urlname = context.getURL("{_HTMLDir_}/forms/WorkItemProperties.html");
    ResponsePage page = null;

    try
    {
      StringBuffer answer = context.loadTemplateFile(urlname, subst);
      page = new ResponsePage(answer);
    }
    catch (IOException xcpt)
    {
      context.log("Template file " + urlname + " could not be read: " + Util.toString(xcpt), true);
      throw new ClientException(context, context.get("Error.HTMLTemplateNotFound", urlname));
    }

    return page;
  }

  /**
  * Creates the page that will be displayed when a work item has been
  * checked out.
  * If the work item was started instead of checked out (because the
  * work item could not be checked out and the <code>startAllowed</code>
  * request parameter was set), the work list is shown again. Otherwise,
  * an HTML template will be loaded from
  * <code>programs/<i>ProgramName</i>.html</code> (where <i>ProgramName</i>
  * is the name of the program assigned to the activity). This HTML template
  * file may contain the following substitution variables:
  * <table border="1" cellspacing="0" cellpadding="4">
  *   <tr>
  *     <th>Name</th>
  *     <th>Replacement</th>
  *   </tr>
  *   <tr>
  *     <td>{_HTMLDir_}</td>
  *     <td>The root URI of the servlet's web application, for example,
  *         <code>/MQWFClient</code>.</td>
  *   </tr>
  *   <tr>
  *     <td>{_MQWFServlet_}</td>
  *     <td>The servlet's name, for example,
  *         <code>/MQWFClient/servlet/Main</code>.</td>
  *   </tr>
  *   <tr>
  *     <td>{_Date_}</td>
  *     <td>The current date.</td>
  *   </tr>
  *   <tr>
  *     <td>{_Time_}</td>
  *     <td>The current time.</td>
  *   </tr>
  *   <tr>
  *     <td>{_Name_}</td>
  *     <td>The name of the work item.</td>
  *   </tr>
  *   <tr>
  *     <td>{_Description_}</td>
  *     <td>The description of the work item.</td>
  *   </tr>
  *   <tr>
  *     <td>{_UserID_}</td>
  *     <td>The owner of the work item.</td>
  *   </tr>
  *   <tr>
  *     <td>{_ActionArea_}</td>
  *     <td>A form containing a <b>Submit</b> button which calls
  *         <code>checkInWorkItem</code> with the input data entered and
  *         a <b>Cancel</b> button which calls <code>cancelWorkItem</code>.</td>
  *   </tr>
  * </table>
  * If the HTML template file cannot be found, and
  * <a href="../../../../../../properties.html#generate">GenerateForms</a>
  * is set to true, a form is created from the work item's input container.
  * Submitting this form will check in the work item with the data entered in
  * the generated form.
  * @param context The request context.
  * @return The response page.
  * @exception ClientException if an error occurred
  */
  public ResponsePage checkOutWorkItemResponse(RequestContext context)
                      throws ClientException
  {
    WorkItem workItem = context.getWorkItem();
    ReadOnlyContainer inData = context.getContainer();

    try
    {
      if (inData == null)
      {
        // work item has been started, not checked out
        return mostRecentList(context, !autoRefresh);
      }
      else
      {
        String workItemName = workItem.name();
        String workItemOid = workItem.persistentOid();
        Properties subst = new Properties();

        Date now = new Date();
        subst.put("{_Date_}", context.formatDate(now));
        subst.put("{_Time_}", context.formatTime(now));
        subst.put("{_Name_}", workItem.name());
        subst.put("{_Description_}", null2Empty(workItem.description()));
        subst.put("{_UserID_}", context.getUserID());

        StringBuffer actionArea = new StringBuffer(1024);
        actionArea.append("<input type=\"hidden\" name=\"command\"")
                  .append(" value=\"").append(Command.WI_CHECKIN.getCommand())
                  .append("\">\r\n");
        actionArea.append("<input type=\"hidden\" name=\"id\"")
                  .append(" value=\"").append(workItemOid)
                  .append("\">\r\n");
        actionArea.append("<input type=\"submit\" name=\"submitData\"")
                  .append(" value=\"")
                  .append(context.get("Viewer.CheckInWorkItem"))
                  .append("\">\r\n");
        actionArea.append("<input type=\"button\" name=\"cancel\"")
                  .append(" value=\"").append(context.get("Command.Cancel"))
                  .append("\" onClick=\"javascript:location.replace('")
                  .append(Command.WI_CANCEL_CHECKOUT
                                 .getURL(context, workItemOid, null, false))
                  .append("')\">\r\n");

        subst.put("{_ActionArea_}", actionArea.toString());

        String urlName = "{_HTMLDir_}/programs/" + workItem.implementation() + ".html";
        urlName = context.getURL(urlName);
        ResponsePage page = null;

        try
        {
          StringBuffer answer = context.loadTemplateFile(urlName, subst);
          answer = replaceINDataValues(context, answer.toString(), inData);
          page = new ResponsePage(answer);
        }
        catch (IOException xcpt)
        {
          if (!generateForms)
            throw new ClientException(context, context.get("Error.HTMLTemplateNotFound", urlName));

          context.log("Template file " + urlName + " could not be read: " + Util.toString(xcpt), false);

          StringBuffer answer = context.openPage(context.get("Viewer.CheckInWorkItemTitle"), 2048);
          answer.append("<body>\r\n");

          // Create a table that holds all input fields
          answer.append("<h2>")
                .append(context.get("Viewer.CheckInWorkItemHeader", workItemName))
                .append("</h2>\r\n");
          answer.append("<br><br>\r\n");

          answer.append(context.generateForm(Command.WI_CHECKIN, workItemOid,
                                             Command.WI_CANCEL_CHECKOUT, workItemOid,
                                             inData, workItem.outContainer(), null, "0"));

          answer.append("</body></html>");

          page = new ResponsePage(answer);
        }

        return page;
      }
    }
    catch (FmcException xcpt)
    {
      throw new ClientException(context, xcpt);
    }
  }

  /**
  * Creates the page that is displayed when a work item has been checked in.
  * This page calls {@link BuiltinHandler#queryWorkItems BuiltinHandler.queryWorkItems()}
  * to show the work list again.
  * @param context The request context.
  * @return The response page.
  * @exception ClientException if an error occurred
  */
  public ResponsePage checkInWorkItemResponse(RequestContext context)
                      throws ClientException
  {
    return mostRecentList(context, !autoRefresh);
  }

  /**
  * Creates the page that is displayed when a work item checkout has been canceled.
  * This page calls {@link BuiltinHandler#queryWorkItems BuiltinHandler.queryWorkItems()}
  * to show the work list again.
  * @param context The request context.
  * @return The response page.
  * @exception ClientException if an error occurred
  */
  public ResponsePage cancelWorkItemResponse(RequestContext context)
                      throws ClientException
  {
    return mostRecentList(context, !autoRefresh);
  }

  /**
  * Creates the page that is displayed when a work item has been finished.
  * This page calls {@link BuiltinHandler#queryWorkItems BuiltinHandler.queryWorkItems()}
  * to show the work list again.
  * @param context The request context.
  * @return The response page.
  * @exception ClientException if an error occurred
  */
  public ResponsePage finishWorkItemResponse(RequestContext context)
                      throws ClientException
  {
    return mostRecentList(context, !autoRefresh);
  }

  /**
  * Creates the page that is displayed when a work item has been force finished.
  * If the work item was actually force finished, this method simply calls
  * the {@link BuiltinHandler#queryWorkItems BuiltinHandler.queryWorkItems()}
  * method. If the work item was not yet force finished because the necessary
  * data has not yet been specified, it will be force finished first.
  * @param context The request context.
  * @return The response page.
  * @exception ClientException if an error occurred
  */
  public ResponsePage forceFinishWorkItemResponse(RequestContext context)
                      throws ClientException
  {
    if (context.getRequestParameter("submitData", null) == null)
    {
      try
      {
        BuiltinHandler builtin = context.getBuiltinHandler();
        WorkItem workItem = context.getWorkItem();
        builtin.forceFinishWorkItem(context, workItem.persistentOid(),
                                    "false", null); // false = w/o container
      }
      catch (FmcException xcpt)
      {
        throw new ClientException(context, xcpt);
      }
    }

    return mostRecentList(context, !autoRefresh);
  }

  /**
  * Creates the page that is displayed when a work item has been started.
  * This page calls {@link BuiltinHandler#queryWorkItems BuiltinHandler.queryWorkItems()}
  * to show the work list again.
  * @param context The request context.
  * @return The response page.
  * @exception ClientException if an error occurred
  */
  public ResponsePage startWorkItemResponse(RequestContext context)
                      throws ClientException
  {
    return mostRecentList(context, !autoRefresh);
  }

  /**
  * Creates the page that is displayed when a work item has been restarted.
  * This page calls {@link BuiltinHandler#queryWorkItems BuiltinHandler.queryWorkItems()}
  * to show the work list again.
  * @param context The request context.
  * @return The response page.
  * @exception ClientException if an error occurred
  */
  public ResponsePage restartWorkItemResponse(RequestContext context)
                      throws ClientException
  {
    return mostRecentList(context, !autoRefresh);
  }

  /**
  * Creates the page that is displayed when a work item has been force restarted.
  * If the work item was actually force restarted, this method simply calls
  * the {@link BuiltinHandler#queryWorkItems BuiltinHandler.queryWorkItems()}
  * method. If the work item was not yet force restarted because the necessary
  * data has not yet been specified, it will be force restarted first.
  * @param context The request context.
  * @return The response page.
  * @exception ClientException if an error occurred
  */
  public ResponsePage forceRestartWorkItemResponse(RequestContext context)
                      throws ClientException
  {
    if (context.getRequestParameter("submitData", null) == null)
    {
      try
      {
        BuiltinHandler builtin = context.getBuiltinHandler();
        WorkItem workItem = context.getWorkItem();
        builtin.forceRestartWorkItem(context, workItem.persistentOid(),
                                     "false", null); // false = w/o container
      }
      catch (FmcException xcpt)
      {
        throw new ClientException(context, xcpt);
      }
    }

    return mostRecentList(context, !autoRefresh);
  }

  /**
  * Creates the page that is displayed when a work item has been terminated.
  * This page calls {@link BuiltinHandler#queryWorkItems BuiltinHandler.queryWorkItems()}
  * to show the work list again.
  * @param context The request context.
  * @return The response page.
  * @exception ClientException if an error occurred
  */
  public ResponsePage terminateWorkItemResponse(RequestContext context)
                      throws ClientException
  {
    return mostRecentList(context, !autoRefresh);
  }

  /**
  * Creates the page that is displayed when an item has been deleted.
  * This page calls {@link BuiltinHandler#queryWorkItems BuiltinHandler.queryWorkItems()}
  * to show the work list again.
  * @param context The request context.
  * @return The response page.
  * @exception ClientException if an error occurred
  */
  public ResponsePage deleteItemResponse(RequestContext context)
                      throws ClientException
  {
    return mostRecentList(context, !autoRefresh);
  }

  /**
  * Creates the page that is displayed when an item has been transferred.
  * If the item was transferred, the work list will be shown again.
  * If the item was not yet transferred because the user ID to transfer
  * to was not specified, the HTML template file <code>forms/Transfer.html</code>
  * will be loaded which may contain the following substitution variables:
  * <table border="1" cellspacing="0" cellpadding="4">
  *   <tr>
  *     <th>Name</th>
  *     <th>Replacement</th>
  *   </tr>
  *   <tr>
  *     <td>{_HTMLDir_}</td>
  *     <td>The root URI of the servlet's web application, for example,
  *         <code>/MQWFClient</code>.</td>
  *   </tr>
  *   <tr>
  *     <td>{_MQWFServlet_}</td>
  *     <td>The servlet's name, for example,
  *         <code>/MQWFClient/servlet/Main</code>.</td>
  *   </tr>
  *   <tr>
  *     <td>{_Oid_}</td>
  *     <td>The OID of the item that is to be transferred.</td>
  *   </tr>
  *   <tr>
  *     <td>{_ItemName_}</td>
  *     <td>The name of the item that is to be transferred.</td>
  *   </tr>
  * </table>
  * @param context The request context.
  * @return The response page.
  * @exception ClientException if an error occurred
  */
  public ResponsePage transferItemResponse(RequestContext context)
                      throws ClientException
  {
    Item item = context.getItem();

    if (context.getRequestParameter("userID", null) != null)
    {
      // transfer successful
      return mostRecentList(context, !autoRefresh);
    }
    else
    {
      // userID needs to be supplied
      Properties subst = new Properties();

      try
      {
        subst.put("{_ItemName_}", item.name());
        subst.put("{_Oid_}", item.persistentOid());
      }
      catch (FmcException xcpt)
      {
        throw new ClientException(context, xcpt);
      }

      String urlName = context.getURL("{_HTMLDir_}/forms/Transfer.html");

      try
      {
        StringBuffer answer = context.loadTemplateFile(urlName, subst);
        return new ResponsePage(answer);
      }
      catch (IOException xcpt)
      {
        context.log("Template file " + urlName + " could not be read: " + Util.toString(xcpt), true);
        throw new ClientException(context, context.get("Error.HTMLTemplateNotFound", urlName));
      }
    }
  }

  /**
  * @deprecated - this function will be redesigned in a later release
  */
  public ResponsePage getFilterAssistantResponse(RequestContext context)
                      throws ClientException
  {
    return null;
  }

  /**
  * @deprecated - this function will be redesigned in a later release
  */
  public ResponsePage setUserSettingsResponse(RequestContext context)
                      throws ClientException
  {
    return new ResponsePage(context.getRedirect(context.getCommand("getUserInfo")));
  }

  /**
  * @deprecated - this function will be redesigned in a later release
  */
  public ResponsePage setViewSettingsResponse(RequestContext context)
                      throws ClientException
  {
    return new ResponsePage(context.getRedirect(context.getCommand("showSettings")));
  }

  /**
  * @deprecated - this function will be redesigned in a later release
  */
  public ResponsePage getUserInfoResponse(RequestContext context)
                      throws ClientException
  {
    return null;
  }

  /**
  * @deprecated - this function will be redesigned in a later release
  */
  public ResponsePage getSystemInfoResponse(RequestContext context)
                      throws ClientException
  {
    return null;
  }

  /**
  * @deprecated - this function will be redesigned in a later release
  */
  protected final static String null2Empty(String str)
  {
    return Util.null2Default(str, "");
  }

  /**
  * @deprecated - this function will be redesigned in a later release
  */
  protected ResponsePage redirectResponse(RequestContext context, String urlname,
                                          Command command, String id, String addl)
  {
    String commandString = command.getCommand();

    if (commandString.equals(context.getRequest().getParameter("command")))
    {
      if (urlname == null)
        urlname = context.getCommand(commandString, id, addl);

      return new ResponsePage((RequestContext)null, urlname);
    }

    if (commandString.startsWith("query"))
    {
      if (addl == null)
        addl = "";

      addl = addl + "&useCache=true";
    }

    return new ResponsePage((RequestContext) null, context.getCommand(commandString, id, addl));
  }

  /**
  * @deprecated - this function will be redesigned in a later release
  */
  protected ResponsePage mostRecentList(RequestContext context,
                                        boolean useCache)
                         throws ClientException
  {
    int kindOfList = context.mostRecentList(enableWorklists, enableInstancelists,
                                            enableTemplatelists, enableLists, useCache);

    switch (kindOfList)
    {
      case SessionContext.TEMPLATELIST:
        return queryProcessTemplatesResponse(context);

      case SessionContext.INSTANCELIST:
        return queryProcessInstancesResponse(context);

      case SessionContext.WORKLIST:
        return queryWorkItemsResponse(context);

      case SessionContext.LISTS:
        return queryListsResponse(context);
      
      case SessionContext.ACTIVITYINSTANCELIST:
        return queryActivityInstancesResponse(context);
    }

    return null;
  }
  
  public ResponsePage queryActivityInstancesResponse(RequestContext context) throws
  ClientException
  {
  	// should never be called!!!! But is necessary at compile time. 
  	if (context.isDebugEnabled())
  	{
  		context.log("DefaultViewer.queryActivityInstancesResponse called. This should NEVER happen!", false);
  	}
  	return null;
  }
  

  /**
  * @deprecated - this function will be redesigned in a later release
  */
  public static void getContainerData(RequestContext context, StringBuffer answer,
                                      String contType, String structName,
                                      ContainerElement[] elements,
                                      Container compData)
  {
    try
    {
      for (int i = 0; elements != null && i < elements.length; ++i)
      {
        ContainerElement element = elements[i];
        //String itemname = element.name(); -> use fullName to automatically receive the qualified names
        String itemname = element.fullName();
        
        // build qualified item name
        //if (structName.length() > 0)
        //   itemname = structName + "." + itemname;

        if (!element.isArray())
          getElementData(context, answer, contType, itemname, element, compData);
        else
        {
          ContainerElement[] elems = element.arrayElements();

          for (int j = 0; elems != null && j < elems.length; ++j)
          {
            //getElementData(context, answer, contType,
            //               itemname + "[" + j + "]", elems[j], compData);
            getElementData(context,answer,contType,elems[j].fullName(),elems[j],compData);
          }
        }
      }
    }
    catch (FmcException ignored)
    {
    }
  }

  /**
  * @deprecated - this function will be redesigned in a later release
  */
  private static void getElementData(RequestContext context, StringBuffer answer, String contType,
                                     String itemname, ContainerElement element,
                                     Container compData)
  {
    try
    {
      if (element.isStruct())
      {
        context.log("Element " + itemname + " is struct -> calling getContainerData recursively", false);
        // recursive call to process all structure elements
        getContainerData(context, answer, contType, itemname,
                         element.structMembers(), compData);
      }
      else
      {
        if (contType.equals("OUT"))
        {
          boolean existsInInputContainer = false;

          try
          {
            if (compData != null)
              existsInInputContainer = compData.getElement(itemname) != null;
          }
          catch (FmcException ignored)
          {
          }
          finally
          {
            if (!existsInInputContainer)
            {
              answer.append("<tr>");
              answer.append("<td>").append(itemname).append("</td>");
              answer.append("<td>").append(element.type()).append("</td>");
              answer.append("<td>");

              if ("BINARY".equals(element.type()))
              {
                answer.append(context.getMemberValue(element, ""));
              }
              else
              {
                answer.append("<input type=\"text\" name=\"")
                      .append(itemname).append("\"")
                      .append(" value=\"")
                      .append(context.getMemberValue(element, ""));

                if (element.type().equals("STRING"))
                  answer.append("\" size=\"60\">");
                else
                  answer.append("\" size=\"15\">");
              }

              answer.append("</td></tr>\r\n");
            }
          }
        }
        else
        {
          // contType == "IN"
          answer.append("<tr>");
          answer.append("<td>").append(itemname).append("</td>");
          answer.append("<td>").append(element.type()).append("</td>");
          String itemvalue = context.getMemberValue(element, null);
          boolean written = false;

          // if this item exists in the output container then create a field else
          // just write text value
          try
          {
            // @EF what if types are different? provide default mapping anyway?
            String compvalue = null;

            try
            {
              if (compData != null)
              {
                ContainerElement from = compData.getElement(itemname);

                if (from != null)
                  compvalue = context.getMemberValue(from, "");
              }
            }
            catch (FmcException ignored)
            {
            }

            if (compvalue != null && !"BINARY".equals(element.type()))
            {
              answer.append("<td><input type=\"text\" name=\"")
                    .append(itemname).append("\"")
                    .append(" value=\"")
                    .append(RequestContext.toHTMLText(itemvalue, compvalue));

              if (element.type().equals("STRING"))
                answer.append("\" size=\"60\"></td>");
              else
                answer.append("\" size=\"15\"></td>");

              written = true;
            }
          }
          catch (FmcException ignored)
          {
          }
          finally
          {
            if (!written)
            {
              answer.append("<td>")
                    .append(RequestContext.toHTMLText(itemvalue, "&nbsp;"))
                    .append("</td>");
            }
          }
          answer.append("</tr>\r\n");
        }
      }
    }
    catch (FmcException ignored) { }
  }

  /**
  * @deprecated - this function will be redesigned in a later release
  */
  protected StringBuffer replaceINDataValues(RequestContext context,
                                             String contents,
                                             ReadOnlyContainer inData)
  {
    StringBuffer newContents = new StringBuffer (1000);
    int beginPos = 0;
    int endPos = 0;
    int idx = 0;

    // Replace all input fields with input data values
    while((beginPos = contents.indexOf("{IN.", endPos)) > 0)
    {
      // Replace the data field with the actual value from input container
      endPos = contents.indexOf('}', beginPos);
      // Get the input item name
      String itemName = contents.substring(beginPos + 4, endPos);
      // Get corresponding output container element
      newContents.append(contents.substring(idx, beginPos));
      try
      {
        String itemValue = context.getMemberValue(inData, itemName, "&nbsp;");
        // Replace with item value
        newContents.append(itemValue);
      }
      catch (FmcException xcpt)
      {
        // If there is a problem, replace the placeholder with the error message
        newContents.append(xcpt.getMessage());
      }

      idx = endPos + 1;
    } // end of while

    // Append rest of the buffer to the new buffer
    newContents.append(contents.substring(idx));

    return newContents;
  }

  /**
  * @deprecated - this function will be redesigned in a later release
  */
  public Vector createNavigatorEntries(RequestContext context, boolean enableLogoff)
                throws ClientException
  {
    Vector navigatorEntries = new Vector();
    BuiltinHandler builtin = context.getBuiltinHandler();
    String queryParam = autoRefresh ? "" : "&useCache=true";

    try
    {
      if (enableWorklists)
      {
        Enumeration worklists = context.getWorkLists();

        if (!worklists.hasMoreElements())
          worklists = builtin.queryWorkLists(context).getWorkLists();

        WorkList workList = context.getWorkList();
        String worklistName = workList == null ? null : workList.name();

        while (worklists.hasMoreElements())
        {
          WorkList list = (WorkList) worklists.nextElement();
          String name = list.name();

          if (name != null && !name.equals(worklistName))
          {
            name = context.get("Navigator.Worklist", name);
            String command = Command.WL_QUERY_WORK_ITEMS.withParameter(queryParam)
                                .getURL(context, list.persistentOid(), null, false);
            navigatorEntries.addElement(new NavigatorEntry(name, command));
          }
        }
      }

      if (enableTemplatelists)
      {
        Enumeration templatelists = context.getTemplateLists();

        if (!templatelists.hasMoreElements())
          templatelists = builtin.queryProcessTemplateLists(context).getTemplateLists();

        ProcessTemplateList templateList = context.getTemplateList();
        String templatelistName = templateList == null ? null : templateList.name();

        while (templatelists.hasMoreElements())
        {
          ProcessTemplateList list = (ProcessTemplateList) templatelists.nextElement();
          String name = list.name();

          if (name != null && !name.equals(templatelistName))
          {
            name = context.get("Navigator.Templatelist", name);
            String command = Command.PTL_QUERY_TEMPLATES.withParameter(queryParam)
                                .getURL(context, list.persistentOid(), null, false);
            navigatorEntries.addElement(new NavigatorEntry(name, command));
          }
        }
      }

      if (enableInstancelists)
      {
        Enumeration instancelists = context.getInstanceLists();

        if (!instancelists.hasMoreElements())
          instancelists = builtin.queryProcessInstanceLists(context).getInstanceLists();

        ProcessInstanceList instanceList = context.getInstanceList();
        String instancelistName = instanceList == null ? null : instanceList.name();

        while (instancelists.hasMoreElements())
        {
          ProcessInstanceList list = (ProcessInstanceList) instancelists.nextElement();
          String name = list.name();

          if (name != null && !name.equals(instancelistName))
          {
            name = context.get("Navigator.Instancelist", name);
            String command = Command.PIL_QUERY_INSTANCES.withParameter(queryParam)
                                .getURL(context, list.persistentOid(), null, false);
            navigatorEntries.addElement(new NavigatorEntry(name, command));
          }
        }
      }
    }
    catch (FmcException xcpt)
    {
      throw new ClientException(context, xcpt);
    }

    if (enableLogoff)
    {
      navigatorEntries.addElement(
         new NavigatorEntry(context.get("Navigator.Logoff"),
                            context.getCommand("logoff")));
    }

    return navigatorEntries;
  }

  /**
  * @deprecated - this function will be redesigned in a later release
  */
  protected static String createNavigatorComboBox(SessionContext ctx,
                                                  Vector navigatorEntries,
                                                  String user)
  {
    StringBuffer comboBox = new StringBuffer(1024);

    comboBox.append("<table width=\"100%\"><tr><td align=left>").append(ctx.get("Navigator.Title")).append(" &nbsp;");
    comboBox.append("<select name=\"Navigator\" size=\"1\" onChange=");
    comboBox.append("\"var v=options[selectedIndex].value; if (v!=null) location=v;\"\r\n");
    comboBox.append("style=\"background-color: rgb(192,192,192); color: rgb(0,0,128)\">\r\n");
    comboBox.append("<option selected>").append(ctx.get("Navigator.WhatNext")).append("</option>\r\n");

    for (int i = 0; i < navigatorEntries.size(); ++i)
    {
      NavigatorEntry entry = (NavigatorEntry)navigatorEntries.elementAt(i);
      comboBox.append("<option value=\"").append(entry.getCommand()).append("\">");
      comboBox.append(entry.getName()).append("</option>\r\n");
    }
    comboBox.append("</select></td><td align=right>").append(user).append("</td></tr></table>");;
    comboBox.append("<hr>");

    return comboBox.toString();
  }

  //----------------------------------------------------------------------
  protected boolean enableTemplatelists = true;
  protected boolean enableInstancelists = true;
  protected boolean enableWorklists = true;
  protected boolean enableLists = true;

  private String  pageAfterLogoff;
  private boolean autoRefresh = false;
  private boolean generateForms = true;
}
