/*BEGIN CMVC */
//*************************************************************************
//
//  Workfile: servlet/sample/EmailHandler.java, v3-java, fmv3
//  Last update: 05/05/25 16:15:40
//  SCCS path, id: /home/flowmark/vc/0/3/5/5/s.39 1.9
//
//************************************************************************/
/*END CMVC */
/*BEGIN COPYRIGHT */
//*****************************************************************************
//
//  Licensed Materials - Property of IBM
//  5697-FM3 5655-A96
//  (C) Copyright IBM Corp. 1999, 2008
//  US Government Users Restricted Rights - Use, duplication or
//  disclosure restricted by GSA ADP Schedule Contract
//  with IBM Corp.
//
//****************************************************************************/
//END COPYRIGHT */
/*BEGIN DISCLAIMER */
//*************************************************************************
//  DISCLAIMER
//  ----------
//  This material contains programming source code for your consideration.
//  These examples have not been thoroughly tested under all conditions.
//  IBM, therefore, cannot guarantee or imply reliability, serviceability,
//  or function of these programs.
//  ALL PROGRAMS CONTAINED HEREIN ARE PROVIDED TO YOU "AS IS", WITHOUT ANY
//  WARRANTIES (EXPRESS OR IMPLIED) OR SUPPORT WHATSOEVER, INCLUDING BUT
//  NOT LIMITED TO ANY IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS
//  FOR A PARTICULAR PURPOSE.
//************************************************************************/
/*END DISCLAIMER */

package com.ibm.workflow.servlet.sample;

import java.util.*;
import java.net.*;

import javax.servlet.http.*;
import javax.activation.*;
import javax.mail.internet.*;
import javax.mail.*;

import com.ibm.workflow.api.*;
import com.ibm.workflow.servlet.client.*;

/**
* A custom handler which will send a notification email to the receiver of a
* transferred item. The handler (which has to be registered as one of the
* active custom handlers) is activated when an SMTP host is configured.
* The ListViewer.jsp will then direct item transfer requests to this
* handler instead of the BuiltinHandler. The EmailHandler performs the same
* task as the BuiltinHandler (i.e. transfers the item), but in addition
* -- if an email-address was specified in the transfer dialog -- it sends a
* notification email. This email (which can be formatted as HTML) by the
* EmailViewer.jsp can contains a href-link so that the receiver can start the
* item immediately from his (HTML-aware) email client.<br>
* <h3>Motivation</h3>
* <p>
* The <b>MIBM WebSphere MQ Workflow Web Client</b> provides a means that a user can
* participate in workflows where all he needs is a web browser and the
* knowledge of some 'logon' web page to point it to. While this usage pattern
* is adequate for users which participate in workflows on a regular basis, it
* is cumbersome to those users which are involved in a workflow infrequently.
* In particular, users who will receive items only as a result of an
* explicit item transfer (e.g. initiated by an administrator or by some
* other user who is authorized to perform such a transfer) will need to be
* notified that work has been assigned to them.</p>
* <p>
* email is a much more appropriate means of delivering items in a
* <i>push</i>-style fashion. To that end a custom handler for the Web Client
* was developed which enhances the built-in <code>"Transfer Item"</code>
* function so that it will optionally send an email notification to the
* recipient of a item. The content of the notification is generated by a
* Java Server Page (<code>EmailViewer.jsp</code>) which -- since the source
* code for it is available to customers -- can be easily customized. As
* provided, the JSP will generate an email which allows the receiver to
* launch the item directly from his email application.</p>
*/
/*
* <h3>Potential Enhancements</h3>
* <ul>
* <li>Currently, the custom handler talks to the SMTP host, the JSP just does
* the email rendering. The information passed to the JSP (see table) is
* limited. Moving the email code into the JSP would allow the handler to use
* JSP redirection which would have the side-effect that the full ViewContext
* would be available to the JSP. I'll experiment with this design.</li>
* <li>Given above redesign, it would be simple to enhance the functionality,
* e.g. to add a <code>'Reject and transfer back'</code> link</li>
* <li>Currently, the handler only transfers the work item, but doesnt do a
* CheckOut. This may annoy the receiver since he may no longer be the owner
* when he clicks the URL. To change this, the handler could invoke the activity
* form, capture the result HTML (including any backend data) and ship it as a
* MIME attachment. A JavaScript could re-launch the form in the client context.
* However, since the form may have pulled in backend data, the user may now work
* on outdated information.</li>
* <li>Currently, the user executing the transfer has to enter the receiver's
* email address manually. Wouldn't it be great if the engine would know (and
* externalize) the user's LDAP identity and the handler could just use JNDI to
* look up his email address?</li>
* <li>Not really an enhancement which makes sense in this context, but the
* mechanics to transform a work item into an email are rather straightforward
* (as long as the activity implementation is realized via a JSP or HTML form).
* How about a UPES or PEA (High-Performance-Java-Bridge) provider, which looks
* up email addresses and ships URL-emails (with or without pulling
* the forms from the servlet)?</li>
* </ul>
*/
public class EmailHandler extends GenericCommandHandler
{
  /**
  * Returns information about the handler, such as author, version,
  * and copyright.
  * @return a String containing handler information
  */
  public String getHandlerInfo()
  {
    return "MQWF Sample - Send notification mail for item transfer V1.1";
  }

  /**
  * Initializes the handler and caches the SMTP host, if configured. This is
  * only done if the prerequisite packages (javax.mail.* and javax.activation.*)
  * could be found.
  * @param config  The servlet's configuration and initialization parameters
  */
  public void init(Config cfg)
  {
    super.init(cfg);

    try
    {
      Class.forName("javax.mail.Message");
      Class.forName("javax.activation.DataHandler");

      smtpHost    = getConfig().getParameter("SMTP", "Host");
      smtpAddress = getConfig().getParameter("SMTP", "Address");
    }
    catch (Throwable xcpt)
    {
      cfg.getContext().log("Error initializing EmailHandler: " + xcpt.toString(), true);
    }
  }

  /**
  * Service entry point to which the 'x-transferItem' requests issued by
  * the ListViewer.jsp are dispatched.
  * @param request  The request the client has made of the servlet
  * @param response The response the servlet sends to the client
  * @return The result page to be sent to the client
  */
  public ResponsePage transferItem(HttpServletRequest request,
                                   HttpServletResponse response)
                      throws ClientException
  {
    SessionContext sc = getSessionContext(request);

    try
    {
      BuiltinHandler  bh = sc.getBuiltinHandler();
      String          id = request.getParameter("id");
      Item            wi = sc.getWorkItem(sc.getWorkListOid(), id); // wi to xfer
      if (wi == null) wi = sc.getProcessInstanceNotification(sc.getWorkListOid(), id);
      if (wi == null) wi = sc.getActivityInstanceNotification(sc.getWorkListOid(), id);
      String          oo = wi.owner();                          // backup old owner
      RequestContext  rc = bh.transferItem(request, response);  // reuse default work item xfer processing

      String toAddress = request.getParameter("email");

      if (smtpHost != null && smtpAddress != null && toAddress.length() > 0) // Host configured, packages available and email address provided
      {
        StringBuffer cmd  = new StringBuffer();
        URL base = new URL(request.getScheme(),
                           // request.getHostName() might be 'localhost'
                           InetAddress.getLocalHost().getHostName(),
                           request.getServerPort(), "");

        // Point to the EmailViewer.jsp, passing the command to be packed as a href-URL (and some other useful data)
        String toUserID = request.getParameter("userID");

        cmd.append(sc.getURL("{_HTMLDir_}/forms/EmailViewer.jsp"))
           .append("?fromUserID=") .append(oo)
           .append("&toUserID=")   .append(toUserID)
           .append("&byUserID=")   .append(sc.getUserID())
           .append("&fromAddress=").append(smtpAddress)
           .append("&toAddress=")  .append(toAddress)
           .append("&systemGroup=").append(sc.getSystemGroup())
           .append("&smtpHost=")   .append(smtpHost)
           .append("&name=")       .append(wi.name())
           .append("&url=")        .append(base.toString() +
                                           sc.getCommand("checkOutWorkItem", id,
                                                         URLEncoder.encode("&userID=" + toUserID)));

        // Resolve and invoke EmailViewer.jsp, capturing the response HTML

        DataHandler dh = new DataHandler(new URL(base, cmd.toString()));

//      java.io.FileOutputStream out = new java.io.FileOutputStream("mail.html");
//      out.write(((String)dh.getContent()).getBytes());
//      out.close();

        // Setup javamail Session, create MIME message, set its attributes and content and send it away.

        Properties p = new Properties();

        p.put("mail.smtp.host", smtpHost);
//      p.put("mail.debug", "true");

        Session s = Session.getDefaultInstance(p, null);

//      s.setDebug(true);

        Message m = new MimeMessage(s);

        m.setFrom(new InternetAddress(smtpAddress));
        m.setRecipients(Message.RecipientType.TO, new InternetAddress[] { new InternetAddress(toAddress) });
        m.setSubject("IBM WebSphere MQ Workflow Workitem Transfer Notification");
        m.setSentDate(new Date());

        try
        {
          m.setContent(dh.getContent(), dh.getContentType());
          Transport.send(m);
        }
        catch (java.net.ConnectException xcpt)
        {
          sc.log("Error reading '" + cmd + "', no mail will be sent", xcpt);
        }
        catch (Exception xcpt)
        {
          sc.log("Error sending '" + cmd + "'", xcpt);
          sc.log("Type = " + dh.getContentType(), false);
          sc.log("Content =\r\n" + dh.getContent(), false);
        }
      }

      return bh.render("transferItem", rc); // reuse default response for work item xfer
    }
    catch (Exception xcpt) { throw new ClientException(request, xcpt); }
  }

  private String smtpHost    = null;
  private String smtpAddress = null;
}
