/*BEGIN CMVC */
//*************************************************************************
//
//  Workfile: WebClient/smp/groupwl/step2/GroupWorkListViewer.sol, v3-java, fmv3
//  Last update: 01/07/12 15:13:47
//  SCCS path, id: /home/flowmark/vc/0/3/6/1/s.27 1.5
//
//************************************************************************/
/*END CMVC */
/*BEGIN COPYRIGHT */
//*****************************************************************************
//
//  Licensed Materials - Property of IBM
//  5697-FM3 5655-A96
//  (C) Copyright IBM Corp. 1999, 2008
//  US Government Users Restricted Rights - Use, duplication or
//  disclosure restricted by GSA ADP Schedule Contract
//  with IBM Corp.
//
//****************************************************************************/
//END COPYRIGHT */
/*BEGIN DISCLAIMER */
//*************************************************************************
//  DISCLAIMER
//  ----------
//  This material contains programming source code for your consideration.
//  These examples have not been thoroughly tested under all conditions.
//  IBM, therefore, cannot guarantee or imply reliability, serviceability,
//  or function of these programs.
//  ALL PROGRAMS CONTAINED HEREIN ARE PROVIDED TO YOU "AS IS", WITHOUT ANY
//  WARRANTIES (EXPRESS OR IMPLIED) OR SUPPORT WHATSOEVER, INCLUDING BUT
//  NOT LIMITED TO ANY IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS
//  FOR A PARTICULAR PURPOSE.
//************************************************************************/
/*END DISCLAIMER */

package com.ibm.workflow.servlet.sample;

import java.util.*;

import com.ibm.workflow.api.*;
import com.ibm.workflow.api.PersistentListPackage.*;
import com.ibm.workflow.servlet.client.*;

/**
* This class displays a restricted work list which can only be left through
* the <code>logoff</code> command. It also ensures that the work list is
* filled through the
* {@link GroupWorkListHandler#queryGroupWorkItems queryGroupWorkItems}
* command instead of the Web Client's builtin <code>queryWorkItems</code>
* command.
*/
public class GroupWorkListViewer extends JSPViewer
{
  /**
  * Called by {@link GroupWorkListHandler} to initialize the viewer.
  * @param config represents the <code>WebClient.properties</code> configuration
  *               file
  */
  public void init(Config config)
  {
    // Make sure the base class is properly initialized for the methods not overriden here
    super.init(config);

    // retrieve the work list names from the configuration file
    workListName = config.getParameter("GroupWorkList", "WorkListName", workListName);
    groupListName = config.getParameter("GroupWorkList", "GroupListName", groupListName);
  }

  /**
  * Displays the initial page after logging on. First the group work list
  * and the users work list are searched and created, if necessary. Then the
  * users work list is filled and displayed.
  * @param context the request context
  * @return a page containing the user's work list
  * @exception com.ibm.workflow.servlet.client.ClientException if an exception
  *                 has occurred in the workflow APIs
  */
  public ResponsePage logonResponse(RequestContext context) throws ClientException
  {
    BuiltinHandler builtin = context.getBuiltinHandler();

    // query and cache all work lists (does not include any work items)
    builtin.queryWorkLists(context);

    // Make sure the group list and the users list exist
    try
    {
      WorkList groupList = null;
      WorkList workList = null;
      Enumeration workLists = context.getWorkLists();

      // first check if the lists already exist
      while (workLists.hasMoreElements())
      {
        WorkList list = (WorkList) workLists.nextElement();
        String name = list.name();

        if (name.equals(workListName))
        {
          workList = list;
          context.log("Found WorkList: " + name + ", filter = '" + workList.filter() + "'",
                      false);
        }
        else if (name.equals(groupListName))
        {
          groupList = list;
          context.log("Found GroupList: " + name + ", filter = '" + groupList.filter() +
                      "', threshold = " + groupList.threshold(), false);
        }
      }

      ExecutionService service = context.getExecutionService();

      // if the users list does not exist, create it
      if (workList == null)
      {
        workList = service.createWorkList(workListName,
                                          TypeOfList.PUBLIC, null, // owner
                                          null,                    // description
                                          "OWNER = CURRENT_USER",  // filter
                                          null,                    // sort criteria
                                          null);                   // threshold
        context.log("Created WorkList: " + workListName + ", filter = '" +
                    workList.filter() + "'", false);
      }

      // if the group list does not exist, create it. note that the
      // maximum number of work items to be transferred to a user is
      // implemented as the threshold of this list
      if (groupList == null)
      {
        groupList = service.createWorkList(groupListName,
                                           TypeOfList.PUBLIC, null, // owner
                                           null,                    // description
                                           null,                    // filter
                                           null,                    // sort criteria
                                           null);                   // threshold
        context.log("Created GroupList: " + groupListName + ", filter = '" +
                    groupList.filter() + "', threshold = " + groupList.threshold(), false);
      }

      // since the worklist and the group list were just created,
      // requery the worklists
      builtin.queryWorkLists(context);

      // fill the work item cache of this user using the Web Client's
      // builtin command handler
      workListOID = workList.persistentOid();
      builtin.queryWorkItems(context, workListOID, false);
    }
    catch (FmcException xcpt)
    {
      throw new ClientException(context, xcpt);
    }

    // use this class to generate the work list page
    return queryWorkItemsResponse(context);
  }

  /**
  * Create a HTML page containing the user's work list.
  * @param context the request context
  * @return a page containing the user's work list
  * @exception com.ibm.workflow.servlet.client.ClientException if an exception
  *                 has occurred in the workflow APIs
  */
  public ResponsePage queryWorkItemsResponse(RequestContext context)
                      throws ClientException
  {
    // pass the work list OIDs to the JSP
    context.setAttribute("workListOID", workListOID);

    // let the JSP do the actual rendering
    return new ResponsePage(context, "/forms/GroupWorkList.jsp");
  }

  /**
  * The name of the users work list.
  * Can be changed in the <code>WebClient.properties</code> file.
  */
  private String workListName = "_WORKLIST";

  /**
  * The OID of the users work list.
  */
  private String workListOID = null;

  /**
  * The name of the group work list.
  * Can be changed in the <code>WebClient.properties</code> file.
  */
  private String groupListName = "_GROUPLIST";
}

