/*BEGIN CMVC */
//*************************************************************************
//
//  Workfile: WebClient/smp/groupwl/step3/GroupWorkListHandler.sol, v3-java, fmv3
//  Last update: 05/05/25 16:28:17
//  SCCS path, id: /home/flowmark/vc/0/3/6/1/s.31 1.5
//
//************************************************************************/
/*END CMVC */
/*BEGIN COPYRIGHT */
//*****************************************************************************
//
//  Licensed Materials - Property of IBM
//  5697-FM3 5655-A96
//  (C) Copyright IBM Corp. 1999, 2008
//  US Government Users Restricted Rights - Use, duplication or
//  disclosure restricted by GSA ADP Schedule Contract
//  with IBM Corp.
//
//****************************************************************************/
//END COPYRIGHT */
/*BEGIN DISCLAIMER */
//*************************************************************************
//  DISCLAIMER
//  ----------
//  This material contains programming source code for your consideration.
//  These examples have not been thoroughly tested under all conditions.
//  IBM, therefore, cannot guarantee or imply reliability, serviceability,
//  or function of these programs.
//  ALL PROGRAMS CONTAINED HEREIN ARE PROVIDED TO YOU "AS IS", WITHOUT ANY
//  WARRANTIES (EXPRESS OR IMPLIED) OR SUPPORT WHATSOEVER, INCLUDING BUT
//  NOT LIMITED TO ANY IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS
//  FOR A PARTICULAR PURPOSE.
//************************************************************************/
/*END DISCLAIMER */

package com.ibm.workflow.servlet.sample;

import javax.servlet.http.*;

import com.ibm.workflow.api.*;
import com.ibm.workflow.servlet.client.*;

/**
* This class adds a new <code>queryGroupWorkItems</code> command to the
* Web Client to provide a so called <i>group work list</i>. A group
* work list is the (potentially large) work list of a virtual user from
* which work items are transferred in small quantities to real user's work
* lists. These users must get their work lists from the
* <code>queryGroupWorkItems</code> command instead of the built in
* <code>queryWorkItems</code> command of the Web Client. This is the
* responsibility of the {@link DefaultViewer} class.
*/
public class GroupWorkListHandler extends GenericCommandHandler
{
  /**
  * Called by the Web Client to initialize the command handler.
  * @param config represents the <code>WebClient.properties</code> configuration
  *               file
  */
  public void init(Config config)
  {
    // never forget to initialize the base class!
    super.init(config);

    // create and initialize our viewer
    groupListViewer = new GroupWorkListViewer(this);
    groupListViewer.init(config);

    // retrieve owner and refresh threshold from the configuration file
    groupListOwner = config.getParameter("GroupWorkList", "GroupListOwner", groupListOwner);

    try
    {
      String num = config.getParameter("GroupWorkList", "RefreshThreshold",
                                       new Integer(minWorkItems).toString());
      minWorkItems = Integer.valueOf(num).intValue();
    }
    catch (NumberFormatException xcpt)
    {
      config.getContext().log("GroupWorkList.RefreshThreshold", xcpt);
    }
  }

  /**
  * Returns information about the command handler, such as author, version,
  * and copyright.
  * @return a String containing handler information
  */
  public String getHandlerInfo()
  {
    return "IBM WebSphere MQ Workflow Group Worklist Handler 1.0";
  }

  /**
  * Provides access to the group work list owner for class
  * {@link DefaultViewer}.
  * @return the owner of the group work list
  */
  public String getGroupListOwner()
  {
    return groupListOwner;
  }

  /**
  * Ensures that all users are restricted to the group work list. The
  * only exception is the group list owner who will get the
  * <code>DefaultViewer</code> as defined in the
  * <code>WebClient.properties</code> file.
  * @param context context of the session that was just created
  */
  public void onLogon(RequestContext context)
  {
    // restrict all users but the group owner to the group work list
    if (!groupListOwner.equals(context.getUserID()))
      context.setViewer(groupListViewer);
  }

  /**
  * Wrapper method to be called from the servlet. It parses the request
  * parameters and passes them on to
  * {@link #queryGroupWorkItems(com.ibm.workflow.servlet.client.RequestContext,
  * java.lang.String, java.lang.String, boolean) queryGroupWorkItems}.
  * @param request  The request the client has made of the servlet
  * @param response The response the servlet sends to the client
  * @return a page containing the user's work list
  * @exception com.ibm.workflow.servlet.client.ClientException if an exception
  *                 has occurred in the workflow APIs
  */
  public ResponsePage queryGroupWorkItems(HttpServletRequest request,
                                          HttpServletResponse response)
                      throws ClientException
  {
    String workListOID = request.getParameter("id");
    String groupListOID = request.getParameter("srcid");
    boolean useCache = booleanValue(request.getParameter("useCache"), true);

    return queryGroupWorkItems(new RequestContext(request),
                               workListOID, groupListOID, useCache);
  }

  /**
  * The actual implementation of the group work list.
  * @param context the request context
  * @param workListOID the OID of the user's work list
  * @param groupListOID the OID of the group's work list
  * @param useCache if <code>true</code> the cached work items will be shown
  *                 unless the minumum {@link #minWorkItems number} of work
  *                 items has been reached. If <code>false</code>, the work
  *                 items are queried from the server.
  * @return a page containing the user's work list
  * @exception com.ibm.workflow.servlet.client.ClientException if an exception
  *                 has occurred in the workflow APIs
  */
  public ResponsePage queryGroupWorkItems(RequestContext context, String workListOID,
                                          String groupListOID, boolean useCache)
                      throws ClientException
  {
    context.log("queryGroupWorkItems(" + useCache + ") called.", false);

    try
    {
      // get the cached list of work items for this user
      WorkItem[] workItems = context.getWorkItems(workListOID);
      context.log(workItems.length + " work items in " + workListOID, false);

      // if it is below the minimum...
      if (workItems.length <= minWorkItems)
      {
        // ... query some work items on the group work list (which has
        // a threshold) ...
        WorkItem[] groupItems = context.getWorkList(groupListOID).queryWorkItems();
        String user = context.getUserID();

        // ... and transfer them to this user
        context.log(groupItems.length + " work items in " + groupListOID +
                    ", transfering " + groupItems.length + " work items", false);

        for (int i = 0; i < groupItems.length; ++i)
        {
          try
          {
            groupItems[i].transfer(user);
          }
          catch (FmcException xcpt)
          {
            // ignore exception if already transferred to another user
            if (xcpt.rc != FmcException.FMC_ERROR_OWNER_ALREADY_ASSIGNED)
              throw xcpt;
          }
        }

        // force querying the new work items
        useCache = false;
      }
    }
    catch (FmcException xcpt)
    {
      throw new ClientException(context, xcpt);
    }

    // use the builtin 'queryWorkItems' command to display the user's
    // work list
    BuiltinHandler builtin = context.getBuiltinHandler();
    builtin.queryWorkItems(context, workListOID, useCache);

    context.log("queryGroupWorkItems done.", false);

    // use our GroupWorkListViewer to create the response HTML page
    return builtin.render("queryGroupWorkItems", context);
  }

  /**
  * Helper method to convert a String to a boolean
  * @param property the String to convert (may be null)
  * @param dflt the default value to return if <code>property</code> is null
  * @returns <code>property</code> converted to a boolean
  */
  private final static boolean booleanValue(String property, boolean dflt)
  {
    return property == null ? dflt : Boolean.valueOf(property).booleanValue();
  }

  /**
  * The <code>Viewer</code> object responsible to create the HTML responses.
  */
  private GroupWorkListViewer groupListViewer;

  /**
  * The ID of the virtual user whose work list is to be shared by a group
  * of real users. Can be changed in the <code>WebClient.properties</code>
  * file.
  */
  private String groupListOwner = "GROUP";

  /**
  * The 'low water mark' of the user's work lists - if it is reached, a
  * couple of new work items will be transfered from the group work list.
  * Can be changed in the <code>WebClient.properties</code> file.
  */
  private int minWorkItems = 5;
}
