<?xml version='1.0'?>
<!-- ================================================================================================================= -->
<!-- This style-sheet processes the XML-data generated by the AuditTrailHandler. The XML input (as produced by the     -->
<!-- AuditTrailHandler.getActivityStatistics() method) contains all checkOut, checkIn and cancelCheckOut events for    -->
<!-- all work items originating from a particular activity in a process template (i.e. the events for the union of all -->
<!-- work items across all process instances). This data is transformed into a report which shows for each involved    -->
<!-- user on how many of those work items he has worked on and what his minimum, maximum and average work time was. It -->
<!-- also demonstrates how a style-sheet can utilize the Bean Scripting Framework (BSF) to include JavaScript's in the -->
<!-- transformation process.                                                                                           -->
<!-- ================================================================================================================= -->
<xsl:stylesheet
 xmlns:xsl     = 'http://www.w3.org/1999/XSL/Transform'
 xmlns:lxslt   = 'http://xml.apache.org/xslt'
 xmlns:java    = 'http://xml.apache.org/xslt/java'
 xmlns:js      = 'ext1'
 version       = '1.0'
 extension-element-prefixes='js'>
 <lxslt:component prefix='js' elements='clearStats' functions='addEvent, cnt, min, max, avg, format'>
  <lxslt:script lang='javascript'>
  <!-- =============================================================================================================== -->
  <!-- JavaScript to aggregate statistics. For each user a set of events objects is maintained which keep track of the -->
  <!-- elapsed time the user has worked on a particular work item. The event objects are then enumerated to calculate  -->
  <!-- the number, minimum, maximum and average work times the user has worked on activities of the requested name.    -->
  <!--                                                                                                                 -->
  <!-- 'users' is the global variable which aggregates the statistical information. It is a dictionary keyed by        -->
  <!-- UserID. Each entry is a 'user' object (defined via the JavaScript prototype below) which in turn holds a        -->
  <!-- dictionary of 'evt' objects (again defined via a JavaScript prototype), keyed by the OID of the work item. Each -->
  <!-- 'addEvent' call by the style-sheet passes the relevant data to the JavaScript which collects the information in -->
  <!-- 'users'. That information is then queried by the style-sheet constructing the summary for each user.            -->
  <!--                                                                                                                 -->
  <!-- As the style-sheet is precompiled by the AuditTrailHandler, the global variable (and the data it refers to)     -->
  <!-- will survive invocations. 'clearStats' is an extension element which the style-sheet calls before aggregating   -->
  <!-- the statistics thus ensuring that old data from a previous invocation is wiped.                                 -->
  <!-- =============================================================================================================== -->

  var users = new java.util.Hashtable();

  function clearStats(xslProcessorContext, elem) { users.clear(); return null; }

  <!-- =============================================================================================================== -->
  <!-- The 'evt' prototype. Its 'add' method sums up the elapsed time between a checkOut and the corresponding         -->
  <!-- cancelCheckOut or checkIn event. The 'ela' method reports the sum; it takes care of the case that a checkIn     -->
  <!-- is pending (by adjusting the sum). The effect of this is that the max/avg times will change if statistics are   -->
  <!-- re-evaluated for an activity for which a checkIn is pending.                                                    -->
  <!-- =============================================================================================================== -->

  var offset = java.util.TimeZone.getDefault().getRawOffset();

  function evt()
  {
    this.sum  = 0;
    this.time = 0;
    this.add  = function(event, time)
    {
      switch (event)
      {
        case 21022: this.time = time; break; // checkOut
        case 21021: // cancelCheckOut
        case 21023: this.sum += time - this.time; this.time = 0; break; // checkIn
      }
    }

    this.ela = function()
    {
      if (this.time != 0) { this.sum += java.lang.System.currentTimeMillis() - this.time - offset; this.time = 0; }

      return this.sum;
    }
  }

  <!-- =============================================================================================================== -->
  <!-- The 'user' prototype. Its 'calc' method (which is called once before query results are returned) sums up the    -->
  <!-- elapsed times for that users 'evt's. It also identifies the minimum and maximum times the user has worked on    -->
  <!-- a work item.                                                                                                    -->
  <!-- =============================================================================================================== -->

  function user()
  {
    this._cnt = 0;
    this._min = 0;
    this._max = 0;
    this._sum = 0;
    this.evt  = new java.util.Hashtable();
    this.cnt  = function() { if (this._cnt == 0) this.calc(); return this._cnt; }
    this.min  = function() { if (this._cnt == 0) this.calc(); return this._min; }
    this.max  = function() { if (this._cnt == 0) this.calc(); return this._max; }
    this.avg  = function() { if (this._cnt == 0) this.calc(); return this._sum / this._cnt; }
    this.calc = function()
    {
      this._cnt = this.evt.size();

      for (var evts = this.evt.elements(); evts.hasMoreElements(); )
      {
        var e = evts.nextElement().ela();

        this._min = this._min == 0 ? e : Math.min(this._min, e);
        this._max = Math.max(this._max, e);
        this._sum += e;
      }
    }
  }

  <!-- =============================================================================================================== -->
  <!-- The 'addEvent' function the style-sheet calls for each AuditEvent. It allocates 'user' and/or 'evt' objects and -->
  <!-- hands over the data provided by the style-sheet.                                                                -->
  <!-- =============================================================================================================== -->

  function addEvent(name, item, event, time)
  {
    var u = users.get(name); if (u == null) { u = new user(); users.put(name, u); }
    var e = u.evt.get(item); if (e == null) { e = new evt (); u.evt.put(item, e); }

    e.add(event, time);

    return null;
  }

  <!-- =============================================================================================================== -->
  <!-- The various statistical query functions, taking the 'UserID' as argument.                                       -->
  <!-- =============================================================================================================== -->

  function cnt(name) { return users.get(name).cnt(); }
  function min(name) { return hours(users.get(name).min()); }
  function max(name) { return hours(users.get(name).max()); }
  function avg(name) { return hours(users.get(name).avg()); }

  <!-- =============================================================================================================== -->
  <!-- Formatting function to convert milliseconds into HH:MM:SS.ss format.                                            -->
  <!-- =============================================================================================================== -->

  function hours(ms)
  {
    var s = ms/1000; var h = Math.floor(s/3600); s -= h*3600; var m = Math.floor(s/60); s -= m*60;

    return h + 'h ' + m + 'm ' + Math.round(s*100)/100 + 's';
  }

  function format(millis) { return new java.sql.Timestamp(parseInt(millis)); } <!-- For detail see AuditTrail.xsl. -->

  </lxslt:script>
 </lxslt:component>
 <!-- =============================================================================================================== -->
 <!-- Formatting parameters for the Navigation bar. Those are passed by the AuditTrailHandler.                        -->
 <!-- =============================================================================================================== -->
 <xsl:param name='css'/>
 <xsl:param name='cmdBack'/>
 <xsl:param name='imgBack'/>
 <xsl:param name='cmdRefresh'/>
 <xsl:param name='imgRefresh'/>
 <xsl:param name='cmdLogoff'/>
 <xsl:param name='imgLogoff'/>
 <!-- ============================================================== -->
 <!-- For the root, produce the Navigation bar and recurse into tree -->
 <!-- ============================================================== -->
 <xsl:output method='html' indent='no'/>
 <xsl:template match='/'>
  <html>
   <head>
    <meta http-equiv='Content-Type' content='text/html; charset=iso-8859-1'/>
    <meta http-equiv='Expires' content='0'/>
    <meta http-equiv='Pragma' content='no-cache'/>
    <meta http-equiv='Cache-Control' content='no-cache'/>
    <title>IBM WebSphere MQ Workflow - Web Client - Activity Statistics</title>
    <link>
     <xsl:attribute name='rel' >stylesheet</xsl:attribute>
     <xsl:attribute name='type'>text/css</xsl:attribute>
     <xsl:attribute name='href'><xsl:value-of select='$css'/></xsl:attribute>
    </link>
   </head>
   <body>
    <table width='100%' border='2' bordercolordark='#000000' bordercolorlight='#FFFFFF' cellpadding='0' cellspacing='0'
           class='navigator'>
     <tr>
      <td align='left'>
       <a>
        <xsl:attribute name='href'><xsl:value-of select='$cmdBack'/></xsl:attribute>
        <img>
         <xsl:attribute name='src'><xsl:value-of select='$imgBack'/></xsl:attribute>
           <xsl:attribute name='alt'>Back to Work List</xsl:attribute>
         <xsl:attribute name='border'>0</xsl:attribute>
        </img>
       </a>
      </td>
      <td>
       <a>
        <xsl:attribute name='href'><xsl:value-of select='$cmdRefresh'/></xsl:attribute>
        <img>
         <xsl:attribute name='src'><xsl:value-of select='$imgRefresh'/></xsl:attribute>
         <xsl:attribute name='alt'>Refresh</xsl:attribute>
         <xsl:attribute name='border'>0</xsl:attribute>
        </img>
       </a>
      </td>
      <td width='100%'>&#160;</td>
      <td align='right'>
       <a>
        <xsl:attribute name='href'><xsl:value-of select='$cmdLogoff'/></xsl:attribute>
        <img>
         <xsl:attribute name='src'><xsl:value-of select='$imgLogoff'/></xsl:attribute>
         <xsl:attribute name='alt'>Log off</xsl:attribute>
         <xsl:attribute name='border'>0</xsl:attribute>
        </img>
       </a>
      </td>
     </tr>
    </table>
    <br/>
    <xsl:choose>
     <xsl:when test='count(AuditTrail/*) &gt; 0'>
      <xsl:apply-templates select='AuditTrail'/>
     </xsl:when>
     <xsl:otherwise>
      <h1>Audit Trail Data unavailable</h1>
      <p><b>No Audit Trail records for this activity have been found. This means that no relevant audit events
      (as defined by the audit trail filter of this activity) have occurred or that processes defining this
      activity are not audited at all.</b></p>
     </xsl:otherwise>
    </xsl:choose>
   </body>
  </html>
 </xsl:template>
 <!-- ================================================================================================================ -->
 <!-- Report common process data from first event, then aggregate, then report summary.                                -->
 <!-- ================================================================================================================ -->
 <xsl:template match='AuditTrail'>
  <p>
   <font face='Arial' size='5'>
    <b>Activity Statistics for Activity <i>&quot;<xsl:value-of select='AuditTrailRecord[1]/ActivityName'/>&quot;</i></b>
   </font>
  </p>
  <table>
   <tr>
    <td><b>Template name</b></td>
    <td><b>:</b></td>
    <td><xsl:value-of select='AuditTrailRecord[1]/ProcTemplName'/></td>
   </tr>
   <tr>
    <td><b>Template valid since</b></td>
    <td><b>:</b></td>
    <td><xsl:value-of select='js:format(string(AuditTrailRecord[1]/ProcTemplValidFromDate))'/></td>
   </tr>
  </table>
  <br/>
  <br/>
  <!-- =============================================================================================================== -->
  <!-- Aggregating 'AuditTrailRecord' records. First old stats are cleared, then the relevant data is passed to the    -->
  <!-- script. Note that the JavaScript relies on the data being sorted by UserID/AssociatedObject/Timestamp.          -->
  <!-- =============================================================================================================== -->
  <js:clearStats/>
  <xsl:for-each select='AuditTrailRecord'>
   <xsl:sort select='UserID' order='ascending'/>
   <xsl:sort select='AssociatedObject' order='ascending'/>
   <xsl:sort select='Timestamp' order='ascending'/>
   <xsl:value-of select='js:addEvent(string(UserID), string(AssociatedObject), number(AuditEvent), number(Timestamp))'/>
  </xsl:for-each>
  <!-- =============================================================================================================== -->
  <!-- Reporting the aggregated statistics in a table, one row per user. The trick is to have the for-each predicate   -->
  <!-- to select the last occurence of each UserID. For this to work the records must be sorted by UserID (which is    -->
  <!-- done by the AuditTrailHandler, i.e. the corresponding SQL-query).                                               -->
  <!-- =============================================================================================================== -->
  <table border='3'>
   <tr bgcolor='#7B94E7'>
    <th>User</th>
    <th>Number of Activities</th>
    <th>Minimum Work Time</th>
    <th>Maximum Work Time</th>
    <th>Average Work Time</th>
   </tr>
   <xsl:for-each select='AuditTrailRecord[string(UserID) != string(following::UserID)]'>
    <tr bgcolor='yellow'>
     <td><xsl:value-of select='UserID'/></td>
     <td align='right'><xsl:value-of select='js:cnt(string(UserID))'/></td>
     <td align='right'><xsl:value-of select='js:min(string(UserID))'/></td>
     <td align='right'><xsl:value-of select='js:max(string(UserID))'/></td>
     <td align='right'><xsl:value-of select='js:avg(string(UserID))'/></td>
    </tr>
   </xsl:for-each>
  </table>
 </xsl:template>
</xsl:stylesheet>
